#! usr/bin/env python3
#  coding=utf-8 #

"""
**This module defines the storage units**

 The storage_units module defines various kinds of storage units with
 associated attributes and methods, from simple to specific ones.

 It includes :
    - StorageUnit : simple storage unit inheriting from EnergyUnit,
      with storage specific attributes. It includes the objective "minimize
      capacity".
    - Thermocline storage : a thermal storage that need to cycle (i.e.
      reach SOC_max) every period of Tcycle

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from pulp import LpBinary

from .energy_units import VariableEnergyUnit
from ...general.optimisation.elements import *

__docformat__ = "restructuredtext en"


class StorageUnit(VariableEnergyUnit):
    """
    **Description**

        Simple Storage unit

    **Attributes**

     * capacity (Quantity): maximal energy that can be stored [kWh]
     * e (Quantity): energy at time t in the storage [kWh]
     * set_soc_min (TechnicalDynamicConstraint): constraining the energy to be
       above the value : soc_min*capacity
     * set_soc_max (TechnicalDynamicConstraint): constraining the energy to be
       below the value : soc_max*capacity
     * pc (Quantity) : charging power [kW]
     * pd (Quantity) : discharging power [kW]
     * uc (Quantity) : binary variable describing the charge of the
       storage unit : 0 : Not charging & 1 : charging
     * calc_e (DefinitionDynamicConstraint) : energy calculation at time t ;
       relation power/energy
     * calc_p (DefinitionDynamicConstraint) : power calculation at time t ;
     power
       flow equals charging power minus discharging power
     * on_off_stor (DefinitionDynamicConstraint) : making u[t] matching with
       storage modes (on/off)
     * def_max_charging (DefinitionDynamicConstraint) : defining the max charging
       power, avoiding charging and discharging at the same time
     * def_max_discharging (DefinitionDynamicConstraint) : defining the max
       discharging power, avoiding charging and discharging at the same time
     * def_min_charging (DefinitionDynamicConstraint) : defining the min charging
       power, avoiding charging and discharging at the same time
     * def_min_discharging (DefinitionDynamicConstraint) : defining the min
       discharging power, avoiding charging and discharging at the same time
     * set_e_0 (ActorConstraint) : set the energy state for t=0
     * e_f (Quantity) : energy in the storage at the end of the time
       horizon, i.e. after the last time step [kWh]
     * e_f_min (TechnicalConstraint) : e_f value is constrained above
       soc_min*capacity
     * e_f_max (TechnicalConstraint) : e_f value is constrained below
       soc_max*capacity
     * set_e_f (ActorConstraint) : when e_f is given, it is set in the
       same way the energy is, but after the last time step
     * calc_e_f (DefinitionConstraint) : when e_f is not given,
       it is calculated in the same way the energy is, but after the last
       time step
     * ef_is_e0 (TechnicalConstraint) : Imposing ef=e0 on the time period.
     * cycles (TechnicalDynamicConstraint) : setting a cycle constraint
       e[t] = e[t+cycles/dt]

    """

    def __init__(self, time, name, pc_min=1e-5, pc_max=1e+5,
                 pd_min=1e-5, pd_max=1e+5, capacity=None, e_0=None,
                 e_f=None, soc_min=0, soc_max=1, eff_c=1, eff_d=1,
                 self_disch=0, self_disch_t=0, ef_is_e0=False, cycles=None,
                 energy_type=None, e_min=-1e+6, e_max=1e+6):
        """
        :param time: TimeUnit describing the studied time period
        :param name: name of the storage unit
        :param pc_min: minimal charging power [kW]
        :param pc_max: maximal charging power [kW]
        :param pd_min: minimal discharging power [kW]
        :param pd_max: maximal discharging power [kW]
        :param capacity: maximal energy that can be stored [kWh]
        :param e_0: initial level of energy [kWh]
        :param e_f: final level of energy [kWh]
        :param soc_min: minimal state of charge [pu]
        :param soc_max: maximal state of charge [pu]
        :param eff_c: charging efficiency
        :param eff_d: discharging efficiency
        :param self_disch: part of the capacity that is self-discharging [pu/h]
        :param self_disch_t: part of the energy that is self-discharging [pu/h]
        :param ef_is_e0: binary describing whether the storage is working at
        constant energy during the entire time period (e_0=e_f) or not.
        :param cycles: number of hours between cycling :e[t] = e[
        t+cycles/dt] [hours]
        :param energy_type: energy type the storage unit is used with
        :param e_min: minimal energy consumed by the storage on the whole
        studied period
        :param e_max: maximal energy consumed by the storage on the whole
        studied period
        """

        VariableEnergyUnit.__init__(self, time, name, flow_direction='in',
                                    p_min=-pd_max, e_min=e_min, e_max=e_max,
                                    p_max=pc_max, energy_type=energy_type)

        # --- Checking the state of charge boundaries of the storage system ---
        if isinstance(soc_min, (float, int)) and isinstance(soc_max, (float,
                                                                      int)):
            if soc_min > soc_max:
                raise ValueError('You cannot have soc_min > soc_max')
            elif soc_min > 1 or soc_min < 0 or soc_max > 1 or soc_max < 0:
                raise ValueError('The soc_min and soc_max values are '
                                 'expressed as percentages of the capacity and'
                                 ' must be set between 0 and 1')
        if soc_min is None or soc_max is None:
            raise ValueError('soc_min and soc_max should not be None')

        self.capacity = Quantity(name='capacity', unit='kWh', value=capacity,
                                 lb=0, vlen=1, parent=self)

        self.e = Quantity(name='e', opt=True,
                          description='energy at t in the storage',
                          unit='kWh', vlen=time.LEN,
                          lb=0, ub=capacity, parent=self)

        self.self_disch_t = self_disch_t

        if isinstance(soc_min, (int, float)):
            self.set_soc_min = TechnicalDynamicConstraint(
                exp_t='{0}_e[t] >= {1} * {0}_capacity'.format(self.name,
                                                              soc_min),
                name='set_soc_min', parent=self)

        elif isinstance(soc_min, list):
            self.set_soc_min = TechnicalDynamicConstraint(
                exp_t='{0}_e[t] >= {1}[t] * {0}_capacity'.format(self.name,
                                                                 soc_min),
                name='set_soc_min', parent=self)

        if isinstance(soc_max, (int, float)):
            self.set_soc_max = TechnicalDynamicConstraint(
                exp_t='{0}_e[t] <= {1} * {0}_capacity'.format(self.name,
                                                              soc_max),
                name='set_soc_max', parent=self)

        elif isinstance(soc_max, list):
            self.set_soc_max = TechnicalDynamicConstraint(
                exp_t='{0}_e[t] <= {1}[t] * {0}_capacity'.format(self.name,
                                                                 soc_max),
                name='set_soc_max', parent=self)

        # --- Charging and discharging powers ---
        self.pc = Quantity(name='pc', opt=True, unit='kW', vlen=time.LEN, lb=0,
                           ub=pc_max,
                           parent=self)
        self.pd = Quantity(name='pd', opt=True, unit='kW', vlen=time.LEN, lb=0,
                           ub=pd_max,
                           parent=self)

        self.uc = Quantity(name="uc", vtype=LpBinary, vlen=time.LEN, opt=True,
                           description='binary variable 0:No charging & '
                                       '1:charging',
                           parent=self)

        # CONSTRAINTS
        # Relation power/energy
        if 0 <= self_disch <= 1 and 0 <= self_disch_t <= 1:
            self.calc_e = \
                DefinitionDynamicConstraint(name='calc_e',
                                            t_range=' for t in time.I[:-1]',
                                            exp_t='{0}_e[t+1] - {0}_e[t]*('
                                                  '1-{1}*time.DT)'
                                                  ' - time.DT * ({0}_pc[t]*{'
                                                  '3}- '
                                                  '{0}_pd[t]*1/{4}- {2}*'
                                                  '{0}_capacity) == 0'
                                            .format(self.name, self_disch_t,
                                                    self_disch,
                                                    eff_c, eff_d), parent=self)

        else:
            raise ValueError('self_disch & self_disch_t should have values '
                             'between 0 and 1 and are set to {0} and {1}'
                             .format(self_disch, self_disch_t))

        # Power flow equals charging power minus discharging power
        self.calc_p = DefinitionDynamicConstraint(
            exp_t='{0}_p[t] == {0}_pc[t] - {0}_pd[t]'.format(self.name),
            t_range='for t in time.I', name='calc_p', parent=self)

        # For storage, as the power can be both positive and negative,
        # an other constraint is needed to make u[t] match with on/off
        self.on_off_stor = DefinitionDynamicConstraint(
            exp_t='{0}_pc[t] + {0}_pd[t] - {0}_u[t] * {eps}'
                  ' >= 0'.format(self.name, eps=0.001),
            t_range='for t in time.I',
            name='on_off_stor', parent=self)

        # Limits to avoid charging and discharging at the same time
        self.def_max_charging = DefinitionDynamicConstraint(
            exp_t='{0}_pc[t] - {0}_uc[t] * {1} <= 0'.format(self.name, pc_max),
            t_range='for t in time.I',
            name='def_max_charging', parent=self)

        self.def_max_discharging = DefinitionDynamicConstraint(
            exp_t='{0}_pd[t] - (1 - {0}_uc[t]) * {1} <= 0'.format(self.name,
                                                                  pd_max),
            t_range='for t in time.I',
            name='def_max_discharging', parent=self)

        self.def_min_charging = DefinitionDynamicConstraint(
            exp_t='{0}_pc[t] - {0}_uc[t] * {1} >= 0'.format(self.name, pc_min),
            t_range='for t in time.I',
            name='def_min_charging', parent=self)

        self.def_min_discharging = DefinitionDynamicConstraint(
            exp_t='{0}_pd[t] + ({0}_uc[t] - {0}_u[t]) * {1} '
                  '>= 0'.format(self.name, pd_min), t_range='for t in time.I',
            name='def_min_discharging', parent=self)

        # --- Constraints for initial and final states of charge ---
        # Setting the state of charge for t=0
        if e_0 is not None:
            self.set_e_0 = ActorConstraint(name='set_e_0',
                                           exp='{0}_e[0] == {1}'
                                           .format(self.name, e_0),
                                           parent=self)

        # e_f should be in between the boundaries
        # [soc_min*capacity; soc_max*capacity]
        # (even when the capacity is not defined), and the result of the
        # last charging/ discharging powers and losses applied the energy at
        # the last timestep.
        if e_f is not None:
            self.e_f = Quantity(name='e_f', opt=False, value=e_f,
                                description='energy in the storage at the '
                                            'end of the time horizon, i.e.'
                                            ' after the last time step',
                                unit='kWh', vlen=1, lb=0, ub=capacity,
                                parent=self)

            if isinstance(soc_min, (int, float)):
                self.e_f_min = DefinitionConstraint(
                    exp='{0}_e_f >= {1} * {0}_capacity'.format(self.name,
                                                               soc_min),
                    name='e_f_min', parent=self)

            elif isinstance(soc_min, list):
                self.e_f_min = DefinitionConstraint(
                    exp='{0}_e_f >= {1}[{2}] * {0}_capacity'.format(
                        self.name, soc_min, time.I[-1]),
                    name='set_soc_min', parent=self)

            if isinstance(soc_max, (int, float)):
                self.e_f_max = DefinitionConstraint(
                    exp='{0}_e_f <= {1} * {0}_capacity'.format(self.name,
                                                               soc_max),
                    name='e_f_max', parent=self)

            elif isinstance(soc_max, list):
                self.e_f_max = DefinitionConstraint(
                    exp='{0}_e_f <= {1}[{2}] * {0}_capacity'.format(
                        self.name, soc_max, time.I[-1]),
                    name='set_soc_max', parent=self)

            self.set_e_f = ActorConstraint \
                (name='set_e_f',
                 exp='{0}_e_f-{0}_e[{1}] == {2}*({0}_pc[{1}]*{3}-'
                     '{0}_pd[{1}]*1/{4}-{5}*{0}_e[{1}]-{6}*{0}_capacity)'
                 .format(self.name, time.I[-1], time.DT, eff_c,
                         eff_d, self_disch_t, self_disch), parent=self)
        else:
            self.e_f = Quantity(name='e_f', opt=True,
                                description='energy in the storage at the end '
                                            'of the time horizon, i.e. after '
                                            'the last time step', unit='kWh',
                                vlen=1, lb=0, ub=capacity, parent=self)
            if isinstance(soc_min, (int, float)):
                self.e_f_min = DefinitionConstraint(
                    exp='{0}_e_f >= {1} * {0}_capacity'.format(self.name,
                                                               soc_min),
                    name='e_f_min', parent=self)

            elif isinstance(soc_min, list):
                self.e_f_min = DefinitionConstraint(
                    exp='{0}_e_f >= {1}[{2}] * {0}_capacity'.format(
                        self.name, soc_min, time.I[-1]),
                    name='set_soc_min', parent=self)

            if isinstance(soc_max, (int, float)):
                self.e_f_max = DefinitionConstraint(
                    exp='{0}_e_f <= {1} * {0}_capacity'.format(self.name,
                                                               soc_max),
                    name='e_f_max', parent=self)

            elif isinstance(soc_max, list):
                self.e_f_max = DefinitionConstraint(
                    exp='{0}_e_f <= {1}[{2}] * {0}_capacity'.format(
                        self.name, soc_max, time.I[-1]),
                    name='set_soc_max', parent=self)

            # e_f calculation
            self.calc_e_f = DefinitionConstraint \
                (name='calc_e_f',
                 exp='{0}_e_f-{0}_e[{1}] == {2}*({0}_pc[{1}]*{3}-'
                     '{0}_pd[{1}]*1/{4}-{5}*{0}_e[{1}]-{6}*{0}_capacity)'
                 .format(self.name, time.I[-1], time.DT, eff_c,
                         eff_d, self_disch_t, self_disch), parent=self)

        # Impose ef_is_e0 on the time period
        if ef_is_e0:
            if e_f is None or e_0 is None or e_f == e_0:
                self.ef_is_e0 = ActorConstraint(
                    exp='{0}_e[0] == {0}_e_f'.format(self.name),
                    name='ef_is_e0', parent=self)
            else:
                raise ValueError('When ef_is_e0 is set to True, e_f OR e_0 '
                                 'should remain set to None')

        if cycles is not None:
            if type(cycles) == int:
                delta_t = int(cycles / time.DT)
                self.set_cycles = TechnicalDynamicConstraint(
                    name='set_cycles',
                    exp_t='{0}_e[t] == {0}_e[t+{1}]'.format(self.name,
                                                            delta_t),
                    t_range='for t in time.I[:-{0}]'.format(delta_t),
                    parent=self)
            else:
                raise TypeError('cycles should be an integer : number of '
                                'hours between cycling (e[t] = e[t+cycles/dt]')

    # OBJECTIVES
    def minimize_capacity(self, weight=1):

        """
        :param weight: Weight coefficient for the objective
        """
        def_capacity = DefinitionDynamicConstraint(
            exp_t='{0}_e[t] <= {0}_capacity'
            .format(self.name),
            t_range='for t in time.I',
            name='def_capacity', parent=self)

        min_capacity = Objective(name='min_capacity',
                                 exp='{0}_capacity'.format(self.name),
                                 weight=weight,
                                 parent=self)

        setattr(self, 'def_capacity', def_capacity)
        setattr(self, 'min_capacity', min_capacity)


class ThermoclineStorage(StorageUnit):
    """
    **Description**

        Class ThermoclineStorage : class defining a thermocline heat storage,
        inheriting from StorageUnit.

    **Attributes**

        * is_soc_max (Quantity) : indicating if the storage is fully charged
          0:No 1:Yes
        * def_is_soc_max_inf (DynamicConstraint) : setting the right value
          for is_soc_max
        * def_is_soc_max_sup (DynamicConstraint) : setting the right value
          for is_soc_max
        * force_soc_max (TechnicalDynamicConstraint) : The energy has to be at least
          once at its maximal value during the period Tcycl.
    """

    def __init__(self, time, name, pc_min=1e-5, pc_max=1e+5,
                 pd_min=1e-5, pd_max=1e+5,
                 capacity=None, e_0=None, e_f=None, soc_min=0,
                 soc_max=1, eff_c=1, eff_d=1, self_disch=0, e_min=-1e6,
                 e_max=1e6, Tcycl=120, ef_is_e0=False):
        """
        :param time: TimeUnit describing the studied time period
        :param name: name of the storage unit
        :param pc_min: minimal charging power [kW]
        :param pc_max: maximal charging power [kW]
        :param pd_min: minimal discharging power [kW]
        :param pd_max: maximal discharging power [kW]
        :param capacity: maximal energy that can be stored [kWh]
        :param e_0: initial level of energy [kWh]
        :param e_f: final level of energy [kWh]
        :param soc_min: minimal state of charge [pu]
        :param soc_max: maximal state of charge [pu]
        :param eff_c: charging efficiency
        :param eff_d: discharging efficiency
        :param self_disch: part of the soc that is self-discharging [pu]
        :param Tcycl: period over which the storage is cycling (reaching at
        least once its max state of charge) [hours]
        :param e_min: minimal energy consumed by the storage on the whole
        studied period
        :param e_max: maximal energy consumed by the storage on the whole
        studied period
        :param ef_is_e0: binary describing whether the storage is working at
        constant energy during the entire time period (e_0=e_f) or not.
         """
        StorageUnit.__init__(self, time, name=name, pc_min=pc_min,
                             pc_max=pc_max, pd_min=pd_min, pd_max=pd_max,
                             capacity=capacity, e_0=e_0, e_f=e_f,
                             soc_min=soc_min, soc_max=soc_max, eff_c=eff_c,
                             eff_d=eff_d, self_disch=self_disch,
                             ef_is_e0=ef_is_e0, e_min=e_min, e_max=e_max,
                             energy_type='Heat')

        # DECISION VARIABLES AND PARAMETERS
        # Creation of quantities needed for the Thermocline model
        self.is_soc_max = Quantity(name='is_soc_max', opt=True, vlen=time.LEN,
                                   vtype=LpBinary,
                                   description='indicates if the storage is '
                                               'fully charged 0:No 1:Yes',
                                   parent=self)

        # CONSTRAINTS
        # Thermocline constraints for charge and discharge
        # Set when we are at the maximal state of charge (soc_max) or not
        epsilon = 0.1

        self.def_is_soc_max_inf = DefinitionDynamicConstraint(
            exp_t='{0}_capacity * {0}_is_soc_max[t] >= ({0}_e[t] - '
                  '{0}_capacity + {1})'.format(self.name, epsilon),
            t_range='for t in time.I', name='def_is_soc_max_inf', parent=self)

        self.def_is_soc_max_sup = DefinitionDynamicConstraint(
            exp_t='{0}_capacity * {0}_is_soc_max[t] <= '
                  '{0}_e[t]'.format(self.name), t_range='for t in time.I',
            name='def_is_soc_max_sup', parent=self)

        # The soc has to be at least one time at soc_max during 5 days
        self.force_soc_max = TechnicalDynamicConstraint(
            exp_t='lpSum({0}_is_soc_max[k] for k in range(t-{1}+1, t)) '
                  '>= 1'.format(self.name, round(Tcycl / time.DT)),
            t_range='for t in time.I[{0}:]'.format(round(Tcycl / time.DT)),
            name='force_soc_max', parent=self)



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module includes the following utils for input data management**

 It contains the following methods:
    - select_csv_file_between_dates() : select data in a .csv file
      between two dates
    - read_enedis_data_csv_file() : select and rearrange the data in a .csv
      file of Enedis (the French Distribion System Operator company),
      possibly between two dates

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import pandas as pd
from omegalpes.general.time import TimeUnit
from copy import deepcopy

__docformat__ = "restructuredtext en"


def select_csv_file_between_dates(file_path=None, start='DD/MM/YYYY HH:MM',
                                  end='DD/MM/YYYY HH:MM', v_cols=[],
                                  sep=';'):
    """
    Select data in a .csv file between two dates

    :param file_path: path of the file to rearrange
    :param start: DD MM YYYY HH:00 : first date which should be considered
    :param end: DD MM YYYY HH:00 : last date which should be considered
    :param v_cols: columns which should be considered
    :return: df: a dataframe considering the dates
    """

    # Read CSV file from path and store into dataframe df
    if not v_cols:
        df = pd.read_csv(file_path, sep=sep, usecols=[0, 1], header=0,
                         names=['date', 'value'])
    else:
        df = pd.read_csv(file_path, sep=sep, usecols=range(len(v_cols) + 1),
                         header=0, names=['date'] + v_cols)

    # Ensure that the 'date' column is at the format datetime
    df['date'] = pd.to_datetime(df['date'], dayfirst=True)

    # Set the date as index
    df = df.set_index(['date'])
    df.sort_index()

    # Convert start and end into the format datetime
    start = pd.to_datetime(start, dayfirst=True)
    end = pd.to_datetime(end, dayfirst=True)

    # Select from start to end
    selected_df = df.loc[start: end]

    if not v_cols:
        return selected_df['value']
    else:
        return selected_df.loc[:, v_cols]


def read_enedis_data_csv_file(file_path=None, start=None, end=None):
    """
    Rearrange the Enedis data in cvs file in oder to have a Dataframe of the
    following form

    DD MM YYYY HH:00 ; a
    DD MM YYYY HH:30 ; b
    ...

    :param file_path: path of the file to rearrange
    :param start: DD MM YYYY HH:00 : first date which should be considered
    :param end: DD MM YYYY HH:00 : last date which should be considered
    :return: df_list: the data as a list
    """

    # read the cvs file and collect the data as DD MM YYYY HH:00 ; a ; b
    df_enedis = pd.read_csv(file_path, sep=';', usecols=[0, 1, 2],
                            names=['date', 'value', 'value_last_30min'])
    df_enedis.loc[:, 'date'] = pd.to_datetime(
        df_enedis.loc[:, 'date'], dayfirst=True)

    # Create a DataFrame with the data corresponding to the first 30 minutes
    df_first_30_min = df_enedis[['date', 'value']]

    # Create a DataFrame with the data corresponding to the last 30 minutes
    df_last_30_min = pd.DataFrame(
        {'date': pd.DatetimeIndex(df_enedis['date']) + pd.offsets.Minute(30),
         'value': df_enedis['value_last_30min'].values})

    # Merge both DataFrame and order
    df = pd.concat([df_first_30_min, df_last_30_min])
    # Ensure that the 'date' column is at the format datetime
    df.loc[:, 'date'] = pd.to_datetime(df.loc[:, 'date'], dayfirst=True)
    df = df.set_index(['date'])
    df = df.sort_index()

    if start and end:
        if isinstance(start, str) and isinstance(end, str):
            # Convert start and end into the format datetime
            start = pd.to_datetime(start, dayfirst=True)
            end = pd.to_datetime(end, dayfirst=True)
            # Select from start to end
            df = df.loc[start: end]
        else:
            raise TypeError("start and end values should be entered as string "
                            "'DD/MM/YYYY'")

    # TODO: enable to check the time
    # start = df.head(1).index.values[0]
    # end = df.tail(1).index.values[0]

    df = df['value'].values.tolist()

    return df


def resample_data(input_list=None, dt_origin=1., dt_final=1.,
                  fill_config="ffill", pick_config="mean"):
    """
    Changing data set in a dt_origin time step into data set in a dt_final
    time step
    :param input_list: list to be resample (list or dict)
    :param dt_origin: the time step of the input dataset (in hours)
    :param dt_final: the wanted time step for the output dataset (in hours)
    :param fill_config: choose the configuration of filling when dt_origin >
    dt_final
    by default: ffil, keeping the same data over the time steps
    other way: interpolate, taking into account the time steps)
    :param pick_config: choose the configuration of picking when dt_origin <
    dt_final
    by default: "mean" which determines the mean value of the time steps to
    be reduced
    :return: output_list: resampled list (pandas Series)
    """

    if dt_origin > dt_final:
        time = TimeUnit(periods=len(input_list) + 1, dt=dt_origin,
                        verbose=False)
        copy_input_list = list(deepcopy(input_list))
        copy_input_list.append(input_list[-1])
        copy_input_list = pd.Series(data=copy_input_list, index=time.DATES)
        dt_origin = int(dt_origin * 3600.)
        dt_final = int(dt_final * 3600.)
        if fill_config == "ffill":
            output_list = copy_input_list.resample(
                rule=str(dt_final) + 'S').ffill()
        elif fill_config == "interpolate":
            output_list = copy_input_list.resample(
                rule=str(dt_final) + 'S').interpolate()
        output_list = output_list[0:len(output_list) - 1]
    else:
        time = TimeUnit(periods=len(input_list), dt=dt_origin, verbose=False)
        copy_input_list = deepcopy(input_list)
        copy_input_list = pd.Series(data=copy_input_list, index=time.DATES)
        dt_origin = int(dt_origin * 3600.)
        dt_final = int(dt_final * 3600.)
        if pick_config == "mean":
            output_list = copy_input_list.resample(
                rule=str(dt_final) + 'S').mean()

    return output_list.values




#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module enables to fill the optimization model and formulate it in
LP or MILP based on the package PuLP (LpProblem)**

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import os
import warnings
import glob
import time as pytime

from pulp import LpProblem, LpStatus, LpVariable, lpSum, LpVariableDict
from pulp.solvers import LpSolver

from ..optimisation.core import OptObject
from ..optimisation.elements import *
from ..time import TimeUnit
from ...energy.energy_nodes import EnergyNode
from ...actor.actor import Actor
from ...energy.units.energy_units import AssemblyUnit

__docformat__ = "restructuredtext en"


class OptimisationModel(LpProblem):
    """
    **Description**

        This class includes the optimization model formulated in LP or MILP
        based on the package PuLP (LpProblem)

    """

    def __init__(self, time, name='optimisation_model'):
        """
        :param name: Name of your optimisation model (str)
        """
        LpProblem.__init__(self, name)
        self.verbose = 1
        self.noOverlap = False
        self.time = time
        self._model_units_list = []
        self._model_quantities_list = []
        self._model_constraints_list = []
        self._model_objectives_list = []

    def add_nodes(self, *nodes):
        """
        Add nodes and all connected units to the model
        Check that the time is the same for the model and all the units

        :param nodes: EnergyNode
        """
        for node in nodes:
            if not isinstance(node, EnergyNode):
                raise TypeError(
                    'You have to add nodes from type "EnergyNode".')
            if node not in self._model_units_list:
                self._model_units_list.append(node)
                node.set_power_balance()

            for unit in node.get_connected_energy_units:
                if unit not in self._model_units_list:
                    self._model_units_list.append(unit)

            for unit in self._model_units_list:
                if hasattr(unit, 'time'):
                    if len(self.time.DATES) == len(unit.time.DATES):
                        for i in range(len(self.time.DATES)):
                            if not self.time.DATES[i] == unit.time.DATES[i]:
                                raise ValueError(
                                    'The unit {} does not operate on the '
                                    'same time of the model'.format(
                                        unit.name))
                    else:
                        raise ValueError('The unit {} does not operate on the '
                                         'same time of the model'.format(
                            unit.name))

                self._add_unit_parent(unit)
                self._add_unit_attributes(unit)

                if hasattr(unit, 'costs' or 'incomes' or 'transaction'):
                    #TODO definir
                    unit.eco_balance = Quantity(name='eco_balance', vlen=1,
                                                parent=unit)
                    unit.calc_eco_balance = DefinitionConstraint(
                        name='calc_eco_balance',
                        exp='{}_eco_balance = {}_incomes - {}_costs + {'
                            '}_transaction'.format(unit.name),
                        parent=unit
                    )

        self._add_quantities()
        self._add_objectives(nodes[0].time)
        self._add_constraints(nodes[0].time)

    def add_nodes_and_actors(self, *nodes_or_actors):
        """
        Add nodes, actors and all connected units to the model
        Check that the time is the same for the model and all the units

        :param nodes_or_actors: EnergyNode or Actor type
        """

        for node_or_actor in nodes_or_actors:
            if isinstance(node_or_actor, Actor):
                if node_or_actor not in self._model_units_list:
                    self._model_units_list.append(node_or_actor)

            elif isinstance(node_or_actor, EnergyNode):
                if node_or_actor not in self._model_units_list:
                    self._model_units_list.append(node_or_actor)
                    node_or_actor.set_power_balance()

                for unit in node_or_actor.get_connected_energy_units:
                    if unit not in self._model_units_list:
                        self._model_units_list.append(unit)

            else:
                raise TypeError('You have to add actors or nodes from type '
                                '"Actor" or "EnergyNode".')

            for unit in self._model_units_list:
                if hasattr(unit, 'time'):
                    if len(self.time.DATES) == len(unit.time.DATES):
                        for i in range(len(self.time.DATES)):
                            if not self.time.DATES[i] == unit.time.DATES[i]:
                                raise ValueError(
                                    'The unit {} does not operate on the '
                                    'same time of the model'.format(
                                        unit.name))
                    else:
                        raise ValueError('The unit {} does not operate on the '
                                         'same time of the model'.format(
                            unit.name))

                self._add_unit_parent(unit)
                self._add_unit_attributes(unit)

        self._add_quantities()
        self._add_objectives(self.time)
        self._add_constraints(self.time)

    def _add_unit_parent(self, unit: OptObject) -> None:
        """
        If the parent of the unit is a OptObject, the method adds it to the
        list of units : self._model_units_list
        Else : nothing happens

        :param unit: OptObject whose parent will be added to the list of units
        if also a OptObject
        """

        try:
            parent = getattr(unit, 'parent')
            if isinstance(parent,
                          OptObject) and parent not in self._model_units_list:
                self._model_units_list.append(parent)
        except AttributeError:
            check_if_unit_could_have_parent(unit)

    def _add_unit_attributes(self, unit: OptObject) -> None:
        """
        Adds :
            - The OptObject elements contained in the unit to the list of units
                self._model_units_list
            - The Quantity elements contained in the unit to the list of
                quantities self._model_quantities_list
            - The Constraint elements contained in the unit to the list of
                constraints self._model_constraints_list
            - The Objective elements contained in the unit to the list of
                objectives self._model_objectives_list

        :param unit: OptObject whose attributes (OptObject, Quantity,
        Constraint
        and Objective) will be added to the respective lists
        """
        try:
            for key in list(unit.__dict__.keys()):
                child = getattr(unit, key)
                if isinstance(child,
                              OptObject) and child not in \
                        self._model_units_list:
                    self._model_units_list.append(child)
                elif isinstance(child, Quantity) and child not in \
                        self._model_quantities_list:
                    child.parent = unit
                    self._model_quantities_list.append(child)
                elif isinstance(child, Constraint) and child.active and child \
                        not in self._model_constraints_list:
                    child.parent = unit
                    self._model_constraints_list.append(child)
                elif isinstance(child, Objective) and child not in \
                        self._model_objectives_list:
                    child.parent = unit
                    self._model_objectives_list.append(child)

        except AttributeError:
            pass

    def _add_quantities(self) -> None:
        """
        Adds all quantities as LpVariable to the list self.variables_list
        """

        print('\n--- Adding all variables to the model ---')
        for quantity in self._model_quantities_list:
            q_name = getattr(quantity, 'name')
            val = getattr(quantity, 'value')
            opt = getattr(quantity, 'opt')
            vtyp = getattr(quantity, 'vtype')
            lb, ub = getattr(quantity, 'lb'), getattr(quantity, 'ub')
            try:
                parent = getattr(quantity, 'parent')
                p_name = parent.name
            except AttributeError:
                raise ValueError('Each Quantity object should have a parent')

            new_name = p_name + '_' + q_name

            self._add_quantity(q_name=new_name, q_val=val, q_type=vtyp,
                               q_lb=lb, q_ub=ub, q_opt=opt,
                               parent=quantity.parent)

    def _add_quantity(self, q_name, q_val, q_type, q_lb, q_ub, q_opt,
                      parent=None):
        """
        Adds a quantity as LpVariable to the list self.variables_list
        """
        if self.verbose:
            print('Adding variable : {0}'.format(q_name))

        # Static or dynamic bounds
        if isinstance(q_lb, list):
            lb_cst_exp = q_name + '[t] >= ' + str(q_lb) + '[t]'
            setattr(parent, 'set_lb', DefinitionDynamicConstraint(
                exp_t=lb_cst_exp,
                                                        name='set_lb',
                                                        parent=parent))
            q_lb = min(q_lb)

        if isinstance(q_ub, list):
            ub_cst_exp = q_name + '[t] <= ' + str(q_ub) + '[t]'
            setattr(parent, 'set_ub', DefinitionDynamicConstraint(
                exp_t=ub_cst_exp,
                                                        name='set_ub',
                                                        parent=parent))
            q_ub = max(q_ub)

        # If the values are stored in a dictionary
        if isinstance(q_val, dict):
            if any(i for i in q_opt.values()):
                globals()[q_name] = LpVariable.dict(name=q_name,
                                                    indexs=q_val.keys(),
                                                    lowBound=q_lb,
                                                    upBound=q_ub,
                                                    cat=q_type)

            else:
                globals()[q_name] = q_val

        # If the values are stored in a list
        elif isinstance(q_val, list):
            for ind, opt in enumerate(q_opt):
                if opt:
                    var_name = q_name + '_{0}'.format(ind)
                    globals()[var_name] = LpVariable(name=var_name,
                                                     lowBound=q_lb,
                                                     upBound=q_ub,
                                                     cat=q_type)
                else:
                    if ind == 0:
                        globals()[q_name] = q_val

        # If the values are stored in a int/float
        elif isinstance(q_val, (int, float)):
            if q_opt:
                globals()[q_name] = LpVariable(name=q_name,
                                               lowBound=q_lb,
                                               upBound=q_ub,
                                               cat=q_type)
            else:
                globals()[q_name] = q_val

        else:
            raise TypeError('Value type of the quantity {0} of unit {1} '
                            'is not taken into account'.format(q_name,
                                                               parent.name))

    def _add_constraints(self, time: TimeUnit) -> None:
        """
        Add all constraints to the model

        :param time: TimeUnit
        """
        print('\n--- Adding all constraints to the model ---')
        for cst in self._model_constraints_list:
            cst_name = cst.parent.name + '_' + cst.name
            cst_exp = cst.exp

            # Print the constraint expression
            if self.verbose:
                print('Adding constraint : {0} , exp = {1}'.format(cst_name,
                                                                   cst_exp))

            if isinstance(cst, DynamicConstraint):
                loop_exp = "".join([cst.t_range, ':\n'
                                                 '\ttry:\n'
                                                 '\t\tself += {0}, '
                                                 ''.format(cst.exp_t),
                                    '"{0}_{1}".format(cst_name, t)\n'
                                    '\t\tself.constraints[ "{0}_{1}".format('
                                    'cst_name, t)].cst = cst\n'
                                    '\texcept TypeError:\n'
                                    '\t\twarnings.warn("Possible error")'])
                exec(loop_exp)

            else:
                try:
                    # Add the constraint to the optimization model
                    self += eval(cst_exp), cst_name
                    # self.addConstraint(eval(cst_exp), cst_name)

                    # Add the whole constraint as an attribute of constraint
                    # added in the optimization model
                    self.constraints[cst_name].cst = cst

                except TypeError:
                    warnings.warn('Possible error')

    def _add_objectives(self, time: TimeUnit) -> None:
        """
        Adds all objectives to the model

        :param time: TimeUnit
        """
        print('\n--- Adding all objectives to the model ---')
        objective = 0
        for obj in self._model_objectives_list:
            obj_name = obj.parent.name + '_' + obj.name
            obj_exp = obj.exp
            obj_weight = str(obj.weight)

            if self.verbose:
                print('Adding objective : {0}'.format(obj_name))
            objective += eval(obj_weight + ' * (' + obj_exp + ')')
        self += objective, "obj_tot"

    def update_units(self):
        """
        Updates all units values with optimization results
        """
        var_dict = self.variablesDict()
        for unit in self._model_units_list:
            print("Updating unit : {0}".format(unit.name))
            for key in list(unit.__dict__.keys()):
                quantity = getattr(unit, key)
                if isinstance(quantity, Quantity):
                    q_name = getattr(quantity, 'name')
                    q_opt = getattr(quantity, 'opt')
                    q_value = getattr(quantity, 'value')

                    print("\tQuantity : {0}".format(q_name))

                    if isinstance(q_value, (float, int)) and q_opt:
                        quantity.value = var_dict[
                            "{0}_{1}".format(unit.name, q_name)].varValue

                    elif isinstance(q_value, list):
                        for i, _ in enumerate(q_opt):
                            if quantity.opt[i]:
                                quantity.value[i] = var_dict[
                                    "{0}_{1}_{2}".format(unit.name,
                                                         q_name, i)].varValue

                    elif isinstance(q_value, dict):
                        for i in list(q_value.keys()):
                            try:
                                quantity.value[i] = var_dict[
                                    "{0}_{1}_{2}".format(unit.name,
                                                         q_name, i)].varValue
                            except KeyError:
                                pass

    def solve_and_update(self, solver: LpSolver = None,
                         find_infeasible_cst_set=False) -> None:
        """
        Solves the optimization model and updates all variables values.

        :param solver: Optimization solver
        :type solver: LpSolver
        """
        print("\n - - - - - RUN OPTIMIZATION - - - - - ")
        if solver is None:
            start_time = pytime.time()
            self.solve(solver=solver, use_mps=False)
            print('Resolution duration =', pytime.time() - start_time,
                  'seconds.')
        else:
            self.solve(solver=solver)

        if LpStatus[self.status] == 'Optimal':
            print("\n - - - - - UPDATE RESULTS - - - - - ")
            self.update_units()
        else:
            # warnings.warn("Your optimization failed with status : {
            # }.".format(
            #     LpStatus[self.status]))
            print("\n\n/!\ Your optimization FAILED with status : {} "
                  "/!\.".format(
                LpStatus[self.status]))
            if LpStatus[self.status] == 'Infeasible':
                print('\nIf you want to catch the source of infeasibility:\n'
                      '* Please download LPFICS and use the method '
                      'find_infeasible_constraint_set(your_model)\n'
                      'You can also use according to your needs:\n'
                      '- find_definition_and_actor_infeasible_constraints_set('
                      'your_model)\n'
                      '- find_definition_and_technical_'
                      'infeasible_constraints_set(your_model)')
                print('* If you are a Gurobi user, you can also refer to the '
                      'method "compute_gurobi_IIS()" in '
                      'general\optimation\model.')
                if find_infeasible_cst_set:
                    find_infeasible_cst_set
            if LpStatus[self.status] == 'Not Solved':
                print('You can maybe try with another solver')

    def get_model_constraints_list(self):
        """ Gets constraints of the model """
        return self._model_constraints_list

    def get_model_constraints_name_list(self):
        """ Gets the names of the constraints of the model """
        constraints_name_list = []
        for constraint in self._model_constraints_list:
            new_objective_name = constraint.parent.name + '_' + constraint.name
            constraints_name_list.append(new_objective_name)
        return constraints_name_list

    def get_model_quantities_list(self):
        """ Gets quantities of the model """
        return self._model_quantities_list

    def get_model_quantities_name_list(self):
        """ Gets the names of the quantities of the model """
        quantities_name_list = []
        for quantity in self._model_quantities_list:
            new_objective_name = quantity.parent.name + '_' + quantity.name
            quantities_name_list.append(new_objective_name)
        return quantities_name_list

    def get_model_objectives_list(self):
        """ Gets objectives of the model """
        return self._model_objectives_list

    def get_model_objectives_name_list(self):
        """ Gets the names of the objectives of the model """
        objectives_name_list = []
        for objective in self._model_objectives_list:
            new_objective_name = objective.parent.name + '_' + objective.name
            objectives_name_list.append(new_objective_name)
        return objectives_name_list


def check_if_unit_could_have_parent(unit):
    """ Checks if the unit has an associated parent

    :param unit: unit which parents will be checked
    """
    import gc
    ref = gc.get_referrers(unit)
    attributes = gc.get_referents(unit)[0]

    for parent in ref:
        try:
            parent_name = parent['name']
            if parent_name not in attributes:
                if not isinstance(unit, (AssemblyUnit, Actor)):
                    warnings.warn('The unit {} seems to have as parent {} '
                                  'which was not declared as '
                                  'parent.'.format(unit.name, parent_name))
                elif isinstance(unit, AssemblyUnit):
                    prod_names = [prod.name for prod in unit.prod_units]
                    cons_name = [cons.name for cons in unit.cons_units]
                    rev_name = [rev.name for rev in unit.rev_units]
                    if (parent_name in prod_names) or (parent_name in
                                                       cons_name) or \
                            parent_name in rev_name:
                        pass
                    else:
                        warnings.warn('The unit {} seems to have as '
                                      'parent {} '
                                      'which was not declared as '
                                      'parent.'.format(unit.name, parent_name))
        except (TypeError, KeyError):
            pass


def compute_gurobi_IIS(gurobi_exe_path=r'C:\gurobi810\win64\bin',
                       opt_model=None, MPS_model=None):
    """
    Identifies the constraints in a .ilp file

    :param gurobi_exe_path: Path to the gurobi solver "gurobi_cl.exe"
    :param opt_model: OptimisationModel to whom compute IIS
    :param MPS_model: name of the mps model
    """
    try:
        # Remove all the existing .mps in your directory
        os.system('"del ' + str('*.mps') + '"')
    except:
        pass

    if MPS_model is None:
        # Create the .mps  in order to get the .mps
        if opt_model is None:
            raise ValueError('You should provide either the OptimisationModel '
                             'or the MPS model')
        opt_model.solve()
        opt_model.writeMPS('{}.mps'.format(opt_model.name))
        MPS_file = glob.glob("*.mps")

    else:
        if MPS_model[-4:] == '.mps':
            MPS_file = [MPS_model]
        else:
            MPS_file = [MPS_model + '.mps']

    # Transform the .mps into a .ilp'
    cmd = '"' + gurobi_exe_path + '\gurobi_cl.exe ResultFile=IISresults.ilp ' \
          + MPS_file[0] + '"'
    os.system(cmd)



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module defines the operator_actor and its scope of responsibility**

 One constraint is available :
    - co2_emission_maximum

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
from ..actor import Actor
from ...general.optimisation.elements import *
from ...energy.units.production_units import ProductionUnit


class RegulatorActor(Actor):
    """
    **Description**

        RegulatorActor class inherits from the the basic class Actor. It
        enables one to model an actor who can add constraints on all
        energy units of the study case.

    **Attributes**

     - name : name of the actor
    """

    def __init__(self, name, verbose=True):
        Actor.__init__(self, name=name, verbose=verbose)

        self.description = 'Regulator Actor OptObject'

    def co2_emission_maximum(self, max_co2, time, cst_production_list):
        """
        To create the actor constraint of a maximum of CO2 emission.

        :param max_co2: Minimum of the CO2 emission. May be an int,
            float or a list with the size of the period study
        :param time: period of the study
        :param cst_unit_list: List of units on which the constraint
            will be applied.

        """
        for production_unit in cst_production_list:
            if isinstance(production_unit, ProductionUnit):
                pass
            else:
                raise TypeError('Should be ProductionUnit type for '
                                'co2_emission_maximum constraint')

        production_name_string = ''
        for production_unit in cst_production_list:
            production_name_string += '_{}'.format(
                production_unit.name)

        production_p_string = ''
        for production_unit in cst_production_list[:-1]:
            production_p_string += '{}_co2_emissions[t] + '.format(
                production_unit.name)
        production_p_string += '{}_co2_emissions[t]'.format(
            cst_production_list[-1].name)

        cst_name = 'max_co2_emission{}'.format(production_name_string)

        if isinstance(max_co2, (int, float)):
            exp = '{0} <= {1}'.format(production_p_string, max_co2)
            setattr(self, cst_name, ActorConstraint(exp=exp,
                                                    name=cst_name))

        elif isinstance(max_co2, list):
            if len(max_co2) != time.LEN:
                raise IndexError(
                    "Your maximal power should be the size of the time "
                    "period : {} but equals {}.".format(time.LEN,
                                                        len(max_co2)))
            else:
                exp_t = '{0} <= {1}[t]'.format(production_p_string, max_co2)
                setattr(self, cst_name, ActorDynamicConstraint(exp_t=exp_t,
                                                               name=cst_name,
                                                               t_range='for '
                                                                       't in '
                                                                 'time.I'))
        else:
            raise TypeError('Your maximal power should be an int, '
                            'a float or a list.')


class LocalAuthorities(RegulatorActor):
    """
    **Description**

        LocalAuthorities class inherits from the basic class RegulatorActor. It
        focuses on local Authorities constraints

    """

    def __init__(self, name):
        RegulatorActor.__init__(self, name=name)


class PublicAuthorities(RegulatorActor):
    """
    **Description**

        PublicAuthorities class inherits from the basic class RegulatorActor. It
        focuses on local Authorities constraints

    """

    def __init__(self, name):
        RegulatorActor.__init__(self, name=name)



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module defines inputs and outputs of as poles**

The poles module includes :
 - FlowPole :  this class defines a pole with a directed flow (in or out)
 - EPole : this class define an energy pole

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

__docformat__ = "restructuredtext en"


class FlowPole(dict):
    """
    **Description**

        Interface for basics flux poles
    """

    def __init__(self, flow='flow', direction='in'):
        """

        :param flow: name of the flow
        :type flow: str
        :param direction: direction of the flow : 'in' or 'out'
        :type direction: str
        """
        dict.__init__(self)

        self.flow = flow
        self.direction = direction


class Epole(FlowPole):
    """
    **Description**

        Definition of an energetic pole, power and power
        flow direction convention 'in' or 'out'
    """

    def __init__(self, p, direction, energy_type=None):
        FlowPole.__init__(self, flow='p', direction='direction')
        self.energy_type = energy_type
        self.update({self.flow: p, self.direction: direction})



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
This module contains the exergy assessment routines of OMEGALPES. This module:

1) Determines inlet, outlet or contained exergy of, respectively, any
   ConsumptionUnit, ProductionUnit or StorageUnit.
2) Determines exergy destruction within any EnergyUnit.
2) Recognizes Electrical and Thermal energy.
3) Can calculate exergy for a single unit or for a list of units.
4) Can proceed with only one temperature value or with a list of temperatures.
   4.1. Formulates exergy for one single EnergyUnit and temperature.
   4.2. Formulates timely exergy if one single EnergyUnit and a list of
        temperatures is provided.
   4.3. Formulates exergy for each unit within a list of EnergyUnits if only
        one temperature is provided.
   4.4. Formulates timely exergy for each unit within a list of EnergyUnits if
        a list of temperatures is provided.

The exergy-related classes defined in this module are not physical units. They
are virtual units attached to their energetic counterparts. Consequently,
the exergy and exergy destruction calculated in this module are attached to
the EnergyUnit as a Quantity at the moment of calculating it.

..

    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from omegalpes.energy.energy_types import *
from omegalpes.energy.units.conversion_units import *
from omegalpes.energy.units.production_units import *
from omegalpes.energy.units.consumption_units import *
from omegalpes.energy.units.storage_units import StorageUnit
from omegalpes.general.optimisation.core import OptObject
from omegalpes.general.optimisation.elements import *

__docformat__ = "restructuredtext en"


class ElectricalExergy(OptObject):

    def __init__(self, energy_unit: EnergyUnit or list):
        OptObject.__init__(self, name='elect_exergy',
                           description="Exergy of an electrical flow",
                           verbose=True)

        _check_energy_unit(energy_unit)
        _check_exergy_not_assessed(energy_unit)
        _check_energy_type(energy_unit, elec)
        self._assess_elec_exergy(energy_unit)

    def _assess_elec_exergy(self, energy_unit):

        if isinstance(energy_unit, EnergyUnit):
            energy_unit.exergy = Quantity(name='exergy',
                                          description='Exergy content of the '
                                                      'energy flow interacting'
                                                      ' with this unit',
                                          vlen=energy_unit.time.LEN, unit='kW',
                                          parent=energy_unit)

            if isinstance(energy_unit, StorageUnit):
                "In storage units, the contained exergy is calculated."
                energy_unit.calc_exergy = DefinitionDynamicConstraint(
                    name='calc_exergy',
                    exp_t='{0}_exergy[t] == {0}_e[t]*time.DT'
                        .format(energy_unit.name),
                    t_range='for t in time.I',
                    parent=self)
            else:
                energy_unit.calc_exergy = DefinitionDynamicConstraint(
                    name='calc_exergy',
                    exp_t='{0}_exergy[t] == {0}_p[t]*time.DT'.
                        format(energy_unit.name),
                    t_range='for t in time.I',
                    parent=self)
        elif isinstance(energy_unit, list):
            for energy_unit in energy_unit:
                energy_unit.exergy = Quantity(name='exergy',
                                              description='Exergy content of '
                                                          'the energy flow '
                                                          'interacting with '
                                                          'this unit',
                                              vlen=energy_unit.time.LEN,
                                              unit='kW',
                                              parent=energy_unit)

                if isinstance(energy_unit, StorageUnit):
                    """"In storage units, exergy is understood as contained 
                    exergy."""
                    energy_unit.calc_exergy = DefinitionDynamicConstraint(
                        name='calc_exergy',
                        exp_t='{0}_exergy[t] == {0}_e[t]*time.DT'
                            .format(energy_unit.name),
                        t_range='for t in time.I',
                        parent=self)
                else:
                    energy_unit.calc_exergy = DefinitionDynamicConstraint(
                        name='calc_exergy',
                        exp_t='{0}_exergy[t] == {0}_p[t]*time.DT'.
                            format(energy_unit.name),
                        t_range='for t in time.I',
                        parent=self)


class ThermalExergy(OptObject):

    def __init__(self, energy_unit: EnergyUnit or list,
                 temp_heat: int or float or list,
                 temp_ref=20):
        OptObject.__init__(self, name='thermal_exergy',
                           description="Exergy of a thermal energy flow",
                           verbose=True)

        _check_energy_unit(energy_unit)
        _check_exergy_not_assessed(energy_unit)
        _check_energy_type(energy_unit, thermal)
        self._check_temperature(energy_unit, temp_heat)

        if isinstance(energy_unit, EnergyUnit):
            energy_unit.exergy = Quantity(name='exergy',
                                          description='Exergy content of the energy'
                                                      ' flow interacting with this unit',
                                          vlen=energy_unit.time.LEN, unit='kW',
                                          parent=energy_unit)
            energy_unit.calc_exergy = DefinitionDynamicConstraint(
                name='calc_exergy',
                exp_t='{0}_exergy[t] == {0}_p[t] * '
                                                              '(1 -({1}+273.15)/({1}+273.15))*time.DT'
                .format(
                                                            energy_unit.name,
                                                            temp_ref),
                t_range='for t in time.I',
                parent=self)
        else:
            for energy_unit in energy_unit:
                energy_unit.exergy = Quantity(name='exergy',
                                              description='Exergy content of the energy'
                                                          ' flow interacting with this unit',
                                              vlen=energy_unit.time.LEN,
                                              unit='kW',
                                              parent=energy_unit)
                energy_unit.calc_exergy = DefinitionDynamicConstraint(
                    name='calc_exergy',
                    exp_t='{0}_exergy[t] == {0}_p[t] * '
                          '(1 -({1}+273.15)/({1}+273.15))*time.DT'
                        .format(energy_unit.name, temp_ref),
                    t_range='for t in time.I',
                    parent=self)

        self._assess_th_exergy(energy_unit, temp_ref, temp_heat)

    def _check_temperature(self, energy_unit, temp_heat):
        if temp_heat is None:
            raise ValueError('Please define a temperature to assess thermal '
                             'exergy of the unit named "{0}".'.format(
                energy_unit.name))
        else:
            if not isinstance(temp_heat, (int, float, list)):
                raise TypeError(
                    'Temperatures for thermal exergy assessment should'
                    ' be an int, a float or a list. In your unit named'
                    ' "{0}", temperature is a {1}. Please check it.'.
                        format(energy_unit.name, type(temp_heat)))
            else:
                if isinstance(temp_heat, list):
                    for t_heat in temp_heat:
                        if not isinstance(t_heat, (int, float, list)):
                            raise TypeError('Temperatures for thermal exergy '
                                            'assessment should be an int, a '
                                            'float or a list. In one of your '
                                            'instantiations you entered a temp'
                                            ' that is a {}, please check it.'.
                                            format(type(t_heat)))
                    if isinstance(energy_unit, list):
                        if len(temp_heat) != len(energy_unit) and len(
                                temp_heat) != energy_unit[0].time.LEN:
                            raise IndexError('Wrong length in your temp list. '
                                             '1) To assess thermal exergy of a'
                                             ' list of units with constant '
                                             'temperature, the list of units '
                                             'and the list of temperatures '
                                             'must be the same length. 2) To '
                                             'assess thermal exergy of a list '
                                             'of units with dynamic temp '
                                             'throughout the year, the list of'
                                             ' temps must be the same length '
                                             'of the time unit. 3) If your '
                                             'thermal units have different '
                                             'timely temperature profiles, '
                                             'you have to assess their '
                                             'exergies separately and always '
                                             'satisfying condition #2.')
                    else:
                        if len(temp_heat) != energy_unit.time.LEN:
                            raise IndexError('Please make sure that the list '
                                             'of temperatures that you provide'
                                             ' for the unit named "{0}" has '
                                             'the same length as the time'
                                             'unit.'.format(energy_unit.name))

    def _assess_th_exergy(self, energy_unit, temp_ref, temp_heat=None):

        """ Calculate thermal exergy of a thermal unit """

        if isinstance(temp_heat, list):
            if isinstance(energy_unit, list):
                if len(temp_heat) == len(energy_unit):
                    for e_unit in energy_unit and temp_heat in temp_heat:
                        e_unit.temp_heat = Quantity(
                            name='t_heat',
                            description='Temperature of the thermal flow',
                            lb=0, value=temp_heat, vlen=1, parent=self)
                        if isinstance(e_unit, StorageUnit):
                            e_unit.calc_exergy = DefinitionDynamicConstraint(
                                name='calc_exergy',
                                exp_t='{0}_exergy[t] == {0}_e[t] * '
                                      '(1 -({1}+273.15)/({0}_t_heat+273.15))*'
                                      'time.DT'
                                    .format(e_unit.name, temp_ref),
                                t_range='for t in time.I',
                                parent=self)
                        else:
                            e_unit.calc_exergy = DefinitionDynamicConstraint(
                                name='calc_exergy',
                                exp_t='{0}_exergy[t] == {0}_p[t] * '
                                      '(1 -({1}+273.15)/({0}_t_heat+273.15))*'
                                      'time.DT'
                                    .format(e_unit.name, temp_ref),
                                t_range='for t in time.I',
                                parent=self)
                elif len(temp_heat) == energy_unit[0].time.LEN:
                    for e_unit in energy_unit:
                        e_unit.temp_heat = Quantity(name='t_heat',
                                                    description='Temperature '
                                                                'of heat flow',
                                                    lb=0, value=temp_heat,
                                                    vlen=e_unit.time.LEN,
                                                    # unit=temp_unit,
                                                    parent=self)
                        if isinstance(e_unit, StorageUnit):
                            e_unit.calc_exergy = DefinitionDynamicConstraint(
                                name='calc_exergy',
                                exp_t='{0}_exergy[t] == {0}_e[t] * '
                                      '(1 -({1}+273.15)/({0}_t_heat[t]+273.15))*'
                                      'time.DT'
                                    .format(e_unit.name, temp_ref),
                                t_range='for t in time.I',
                                parent=self)
                        else:
                            e_unit.calc_exergy = DefinitionDynamicConstraint(
                                name='calc_exergy',
                                exp_t='{0}_exergy[t] == {0}_p[t] * '
                                      '(1 -({1}+273.15)/({0}_t_heat[t]+273.15))*'
                                      'time.DT'
                                    .format(e_unit.name, temp_ref),
                                t_range='for t in time.I',
                                parent=self)
            else:
                energy_unit.temp_heat = Quantity(name='t_heat',
                                                 description='Temperature '
                                                             'of heat flow',
                                                 lb=0, value=temp_heat,
                                                 vlen=energy_unit.time.LEN,
                                                 # unit=temp_unit,
                                                 parent=self)
                if isinstance(energy_unit, StorageUnit):
                    energy_unit.calc_exergy = DefinitionDynamicConstraint(
                        name='calc_exergy',
                        exp_t='{0}_exergy[t] == {0}_e[t] * '
                              '(1 -({1}+273.15)/({0}_t_heat[t]+273.15))*time.DT'
                            .format(energy_unit.name, temp_ref),
                        t_range='for t in time.I',
                        parent=self)

                else:
                    energy_unit.calc_exergy = DefinitionDynamicConstraint(
                        name='calc_exergy',
                        exp_t='{0}_exergy[t] == {0}_p[t] * '
                              '(1 -({1}+273.15)/({0}_t_heat[t]+273.15))*time.DT'
                            .format(energy_unit.name, temp_ref),
                        t_range='for t in time.I',
                        parent=self)
        else:
            if isinstance(energy_unit, list):
                for e_unit in energy_unit:
                    e_unit.temp_heat = Quantity(
                        name='t_heat',
                        description='Temperature of the thermal flow',
                        lb=0, value=temp_heat, vlen=1, parent=self)

                    if isinstance(e_unit, StorageUnit):
                        e_unit.calc_exergy = DefinitionDynamicConstraint(
                            name='calc_exergy',
                            exp_t='{0}_exergy[t] == {0}_e[t] * '
                                  '(1 -({1}+273.15)/({2}+273.15))*time.DT'
                                .format(e_unit.name, temp_ref,
                                        e_unit.temp_heat),
                            t_range='for t in time.I',
                            parent=self)
                    else:
                        e_unit.calc_exergy = DefinitionDynamicConstraint(
                            name='calc_exergy',
                            exp_t='{0}_exergy[t] == {0}_p[t] * '
                                  '(1 -({1}+273.15)/({2}+273.15))*time.DT'
                                .format(e_unit.name, temp_ref,
                                        e_unit.temp_heat),
                            t_range='for t in time.I',
                            parent=self)
            else:
                energy_unit.temp_heat = Quantity(
                    name='t_heat',
                    description='Temperature of the thermal flow',
                    lb=0, value=temp_heat, vlen=1, parent=self)
                if isinstance(energy_unit, StorageUnit):
                    energy_unit.calc_exergy = DefinitionDynamicConstraint(
                        name='calc_exergy',
                        exp_t='{0}_exergy[t] == {0}_e[t] * '
                              '(1 -({1}+273.15)/({2}+273.15))*time.DT'
                            .format(energy_unit.name, temp_ref,
                                    energy_unit.temp_heat),
                        t_range='for t in time.I',
                        parent=self)
                else:
                    energy_unit.calc_exergy = DefinitionDynamicConstraint(
                        name='calc_exergy',
                        exp_t='{0}_exergy[t] == {0}_p[t] * '
                              '(1 -({1}+273.15)/({2}+273.15))*time.DT'
                            .format(energy_unit.name, temp_ref,
                                    energy_unit.temp_heat),
                        t_range='for t in time.I',
                        parent=self)


def _check_energy_unit(energy_unit):
    if isinstance(energy_unit, list):
        for energy_unit in energy_unit:
            if not isinstance(energy_unit, EnergyUnit):
                raise TypeError("You are trying to assess exergy on an object "
                                "that is not an EnergyUnit. Please re-check "
                                "your exergy instantiations.")
    else:
        if not isinstance(energy_unit, EnergyUnit):
            raise TypeError("You are trying to assess exergy on an object "
                            "that is not an EnergyUnit. Please re-check "
                            "your exergy instantiations.")


def _check_exergy_not_assessed(energy_unit):
    if isinstance(energy_unit, list):
        for energy_unit in energy_unit:
            if hasattr(energy_unit, 'exergy'):
                raise AttributeError('You are trying to assess {0} exergy more'
                                     ' than once on the unit named "{1}". '
                                     'Please delete all redundant {0} exergy '
                                     'instantiations on that unit.'.format(
                    energy_unit.energy_type,
                    energy_unit.name))
    else:
        if hasattr(energy_unit, 'exergy'):
            raise AttributeError('You are trying to assess {0} exergy more'
                                 ' than once on the unit named "{1}". '
                                 'Please delete all redundant {0} exergy '
                                 'instantiations on that unit.'.format(
                energy_unit.energy_type,
                energy_unit.name))


def _check_energy_type(energy_unit, energy_type):
    if isinstance(energy_unit, list):
        for energy_unit in energy_unit:
            if not energy_unit.energy_type:
                raise TypeError(
                    'Exergy can only be calculated for simple units, '
                    'and not for units made up of other units, such as '
                    'your unit named {0}. Try calculating exergy in '
                    'the subunits of your unit named {0}'.
                        format(energy_unit.name))
            else:
                if energy_unit.energy_type is not energy_type:
                    raise TypeError(
                        '{0} exergy can only be evaluated for {0} units. Your '
                        'unit "{1}" is a {2} unit.'.format(energy_type,
                                                           energy_unit.name,
                                                           energy_unit.
                                                           energy_type))
    else:
        if not energy_unit.energy_type:
            raise TypeError('Exergy can only be calculated for simple units, '
                            'and not for units made up of other units, such as'
                            'your unit named {0}. Try calculating exergy in '
                            'the subunits of your unit named {0}'.
                            format(energy_unit.name))

        else:
            if energy_unit.energy_type is not energy_type:
                raise TypeError('Electrical exergy can only be evaluated for '
                                'electrical units. Your unit named {0} is a '
                                '{1} unit.'.format(energy_unit.name,
                                                   energy_unit.energy_type))


class ExergyDestruction(OptObject):
    def __init__(self, energy_unit=None, exergy_eff=1, temp_ref=20,
                 temp_heat=None):
        OptObject.__init__(self, name='elec_exergy',
                           description="Exergy of an electrical flow",
                           verbose=True)

        if hasattr(energy_unit, 'exergy_dest'):
            raise AttributeError(
                'You have already calculated exergy destruction for the unit '
                '{}'.format(energy_unit.name))

        energy_unit.exergy_dest = Quantity(name='exergy_dest',
                                           description='Exergy destruction',
                                           vlen=energy_unit.time.LEN,
                                           unit='kW', parent=self)
        energy_unit.exd_tot = Quantity(name='exd_tot', unit='kWh', parent=self,
                                       description='total exergy destruction',
                                       vlen=1)
        energy_unit.calc_exd_tot = DefinitionConstraint(name='calc_exd_tot',
                                                        parent=self,
                                                        exp='{0}_exd_tot == '
                                                            'time.DT * '
                                                  'lpSum({0}_exergy_dest[t]'
                                                  ' for t in time.I)'
                                                        .format(
                                                            energy_unit.name))

        """ Calculate exergy destruction within this unit """

        if isinstance(energy_unit, StorageUnit):
            energy_unit.calc_exergy_dest = DefinitionDynamicConstraint(
                name='calc_exergy_dest',
                exp_t='{0}_exergy_dest[t] == {0}_e[t] * {1} * '
                      '(1 -({2}+273.15)/({3}+273.15))*time.DT'
                      .format(energy_unit.name, energy_unit.self_disch_t,
                            temp_ref, temp_heat),
                t_range='for t in time.I',
                parent=self)

        elif isinstance(energy_unit, ConversionUnit):
            exd_conv_unit = energy_unit.name + '_exergy_dest[t] == '
            exergy_balance = ''
            for cons in energy_unit.cons_units:
                if hasattr(cons, 'exergy'):
                    exergy_balance += ' + ' + cons.name + '_exergy[t]'
                else:
                    raise AttributeError(
                        "You need to calculate exergy of subunit "
                        "{0} before calculating exergy destruction of"
                        " the unit named {1}.".format(cons.name,
                                                      energy_unit.name))
            for prod in energy_unit.prod_units:
                if hasattr(prod, 'exergy'):
                    exergy_balance += ' - ' + prod.name + '_exergy[t]'
                else:
                    raise AttributeError(
                        "You need to calculate exergy of subunit "
                        "{0} before calculating exergy destruction of"
                        " the unit named {1}.".format(prod.name,
                                                      energy_unit.name))
            if exergy_balance[0] == ' + ':
                exergy_balance = exergy_balance[1:]
            exd_conv_unit += exergy_balance
            energy_unit.calc_exergy_dest = DefinitionDynamicConstraint(
                name='calc_exergy_dest',
                exp_t=exd_conv_unit,
                t_range='for t in time.I', parent=self)

        elif isinstance(energy_unit, ConsumptionUnit):
            if exergy_eff is None:
                raise ValueError("You should specify an exergy efficiency for "
                                 "unit named {0}.".format(energy_unit.name))

            elif isinstance(exergy_eff, (int, float)):
                energy_unit.calc_exergy_dest = DefinitionDynamicConstraint(
                    name='calc_exergy_dest',
                    exp_t='{0}_exergy_dest[t] == '
                          '{0}_exergy[t] * (1 -{1})'
                        .format(energy_unit.name, exergy_eff),
                    t_range='for t in time.I', parent=self)

            elif isinstance(exergy_eff, list):
                if len(exergy_eff) != energy_unit.time.LEN:
                    raise IndexError('The vector containing exergy '
                                     'efficiency values of the unit named'
                                     ' {0} should be the same length of '
                                     'the time vector.')
                else:
                    energy_unit.calc_exergy_dest = DefinitionDynamicConstraint(
                        name='calc_exergy_dest',
                        exp_t='{0}_exergy_dest[t] == '
                              '{0}_exergy[t] * (1 -{1}[t])'
                            .format(energy_unit.name, exergy_eff),
                        t_range='for t in time.I', parent=self)

            else:
                raise ValueError('Exergy efficiency of the unit named {0} '
                                 'is a {1} but it should be an int, a float'
                                 ' or a list.'.format(energy_unit.name,
                                                      type(exergy_eff)))

        elif isinstance(energy_unit, ProductionUnit):
            if exergy_eff is None:
                raise ValueError("You should specify an exergy efficiency for "
                                 "unit named {0}.".format(energy_unit.name))

            elif isinstance(exergy_eff, (int, float)):
                if exergy_eff == 0:
                    raise ValueError("Exergy efficiency of the unit "
                                     "named {0} has to be greater than 0.".format(
                        energy_unit.name))
                else:
                    energy_unit.calc_exergy_dest = DefinitionDynamicConstraint(
                        name='calc_exergy_dest',
                        exp_t='{0}_exergy_dest[t] == '
                              '{0}_exergy[t] * (1/{1}-1)'
                            .format(energy_unit.name, exergy_eff),
                        t_range='for t in time.I', parent=self)

            elif isinstance(exergy_eff, list):
                if len(exergy_eff) != energy_unit.time.LEN:
                    raise IndexError('The list containing exergy '
                                     'efficiency values of the unit named'
                                     ' "{0}" should be the same length of '
                                     'the time vector.'.
                                     format(energy_unit.name))
                else:
                    for exergy_eff in exergy_eff:
                        if exergy_eff is not (int, float):
                            raise ValueError('In the unit named "{0}", all '
                                             'values in the list of '
                                             'efficiencies must be an int or '
                                             ' float.'.format(energy_unit.
                                                              name))
                        else:
                            if exergy_eff == 0:
                                raise ValueError(
                                    'Exergy efficiency of the unit '
                                    '"{0}" has to be greater than 0 '
                                    'at every time step.'.format(
                                        energy_unit.name))

                    energy_unit.calc_exergy_dest = DefinitionDynamicConstraint(
                        name='calc_exergy_dest',
                        exp_t='{0}_exergy_dest[t] == '
                              '{0}_exergy[t] * (1/{1}[t]-1)'
                            .format(energy_unit.name, exergy_eff),
                        t_range='for t in time.I', parent=self)
            else:
                raise TypeError('Exergy efficiency of the unit named {0} '
                                'is a {1} but it should be an int, a float'
                                ' or a list.'.format(energy_unit.name,
                                                     type(exergy_eff)))

        else:
            raise TypeError('Sorry, the type of the unit named {0}'
                            ' was not recognized for exergy '
                            'analysis.'.format(energy_unit.name))



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module defines the production units**

The production_units module defines various kinds of production units with
associated attributes and methods.

It includes :
 - ProductionUnit : simple production unit inheriting from EnergyUnit and
   with an outer flow direction. The outside co2 emissions, the starting cost,
   the operating cost, the minimal operating time, the minimal non-operating
   time, the maximal increasing ramp and the maximal decreasing ramp can be
   filled.

   Objectives are also available :

    * minimize starting cost, operating cost, total cost
    * minimize production, co2_emissions, time of use
    * maximize production

 - FixedProductionUnit : Production unit with a fixed production profile.
 - VariableProductionUnit : Production unit with a variation of power between
   p_min et p_max.

And also :
 - SeveralProductionUnit: Production unit based on a fixed production curve
   enabling to multiply several times (nb_unit) the same production curve
 - SeveralImaginaryProductionUnit: Production unit based on a fixed
   production curve enabling to multiply several times (nb_unit) the same
   production curve. Be careful, the solution may be imaginary as nb_unit
   can be continuous. The accurate number of the production unit should be
   calculated later
 - SquareProductionUnit: Production unit with a fixed value and fixed
   duration.
 - ShiftableProductionUnit: Production unit with shiftable production
   profile.

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import warnings
from pulp import LpBinary, LpInteger, LpContinuous

from .energy_units import *
del AssemblyUnit
from ...general.optimisation.elements import *
from ...general.optimisation.elements import Objective
from ...general.optimisation.elements import Quantity

__docformat__ = "restructuredtext en"


class ProductionUnit(EnergyUnit):
    """
    **Description**

        Simple Production unit

    **Attributes**

     * co2_out: outside co2 emissions
     * starting_cost: the starting cost
     * operating_cost: the operating cost
     * min_time_on : the minimal operating time
     * min_time_off : the minimal non-operating time
     * max_ramp_up : the  maximal increasing ramp
     * max_ramp_down ; the maximal decreasing ramp
    """

    def __init__(self, time, name, p=None, p_min=1e-5, p_max=1e+5, e_min=0,
                 e_max=1e6, co2_out=None, starting_cost=None,
                 operating_cost=None, min_time_on=None, min_time_off=None,
                 max_ramp_up=None, max_ramp_down=None, availability_hours=None,
                 energy_type=None, verbose=True, no_warn=True):

        EnergyUnit.__init__(self, time, name, flow_direction='out', p=p,
                            p_min=p_min, p_max=p_max, e_min=e_min, e_max=e_max,
                            co2_out=co2_out, starting_cost=starting_cost,
                            operating_cost=operating_cost,
                            min_time_on=min_time_on,
                            min_time_off=min_time_off, max_ramp_up=max_ramp_up,
                            max_ramp_down=max_ramp_down,
                            availability_hours=availability_hours,
                            energy_type=energy_type,
                            verbose=verbose, no_warn=no_warn)

    def minimize_production(self, weight=1):
        """

        :param weight: Weight coefficient for the objective
        """
        self.minimize_energy(weight=weight)
        self.min_energy.name = 'min_production'

    def maximize_production(self, weight=1):
        """

        :param weight: Weight coefficient for the objective
        """
        self.minimize_energy(weight=-1 * weight)
        self.min_energy.name = 'max_production'


class FixedProductionUnit(FixedEnergyUnit, ProductionUnit):
    """
    **Description**

        Production unit with a fixed production profile.

    **Attributes**

     * p : instantaneous power production known by advance (kW)
     * energy_type : type of energy ('Electrical', 'Heat', ...)


    """

    def __init__(self, time, name: str, p: list or dict or pd.DataFrame = None,
                 co2_out=None,
                 starting_cost=None, operating_cost=None, energy_type=None,
                 verbose=True):
        ProductionUnit.__init__(self, time=time, name=name, verbose=verbose)
        FixedEnergyUnit.__init__(self, time, name=name, p=p,
                                 flow_direction='out',
                                 starting_cost=starting_cost,
                                 operating_cost=operating_cost, co2_out=co2_out,
                                 energy_type=energy_type,
                                 verbose=False)


class VariableProductionUnit(VariableEnergyUnit, ProductionUnit):
    """
    **Description**

        Production unit with a variation of power between p_min et p_max.

    **Attributes**

     * p_max : maximal instantaneous power production (kW)
     * p_min : minimal instantaneous power production (kW)
     * energy_type : type of energy ('Electrical', 'Heat', ...)

    """

    def __init__(self, time, name, p_min=1e-5, p_max=1e+5, e_min=0,
                 e_max=1e6, co2_out=None, starting_cost=None,
                 operating_cost=None, min_time_on=None, min_time_off=None,
                 max_ramp_up=None, max_ramp_down=None, energy_type=None,
                 verbose=True, no_warn=True):
        ProductionUnit.__init__(self, time=time, name=name, verbose=verbose)
        VariableEnergyUnit.__init__(self, time, name=name,
                                    flow_direction='out', p_min=p_min,
                                    p_max=p_max, e_min=e_min, e_max=e_max,
                                    starting_cost=starting_cost,
                                    operating_cost=operating_cost,
                                    min_time_on=min_time_on,
                                    min_time_off=min_time_off,
                                    max_ramp_up=max_ramp_up,
                                    max_ramp_down=max_ramp_down,
                                    co2_out=co2_out, energy_type=energy_type,
                                    verbose=False,
                                    no_warn=no_warn)


class SeveralProductionUnit(VariableProductionUnit, SeveralEnergyUnit):
    """
    **Description**

        Production unit based on a fixed production curve enabling to multiply
        several times (nb_unit) the same production curve.
        nb_unit is an integer variable.

    **Attributes**

     * fixed_prod : fixed production curve

    """

    def __init__(self, time, name, fixed_prod, p_min=1e-5, p_max=1e+5, e_min=0,
                 e_max=1e6, nb_unit_min=0, nb_unit_max=None, co2_out=None,
                 starting_cost=None, operating_cost=None, max_ramp_up=None,
                 max_ramp_down=None, energy_type=None,
                 verbose=True, no_warn=True):
        VariableProductionUnit.__init__(self, time=time, name=name,
                                        verbose=verbose)
        SeveralEnergyUnit.__init__(self, time, name=name,
                                   fixed_power=fixed_prod, p_min=p_min,
                                   p_max=p_max, imaginary=False, e_min=e_min,
                                   e_max=e_max, nb_unit_min=nb_unit_min,
                                   nb_unit_max=nb_unit_max,
                                   flow_direction='out',
                                   starting_cost=starting_cost,
                                   operating_cost=operating_cost,
                                   max_ramp_up=max_ramp_up,
                                   max_ramp_down=max_ramp_down,
                                   co2_out=co2_out, energy_type=energy_type,
                                   verbose=False,
                                   no_warn=no_warn)


class SeveralImaginaryProductionUnit(VariableProductionUnit, SeveralEnergyUnit):
    """
    **Description**

        Production unit based on a fixed production curve enabling to multiply
        several times (nb_unit) the same production curve.
        Be careful, the solution may be imaginary as nb_unit can be
        continuous. The accurate number of the production unit should be
        calculated later.

    **Attributes**

     * fixed_prod : fixed production curve

    """

    def __init__(self, time, name, fixed_prod, p_min=1e-5, p_max=1e+5, e_min=0,
                 e_max=1e6, nb_unit_min=0, nb_unit_max=None, co2_out=None,
                 starting_cost=None, operating_cost=None, max_ramp_up=None,
                 max_ramp_down=None, energy_type=None,
                 verbose=True, no_warn=True):
        VariableProductionUnit.__init__(self, time=time, name=name,
                                        verbose=verbose)
        SeveralEnergyUnit.__init__(self, time, name=name,
                                   fixed_power=fixed_prod, p_min=p_min,
                                   p_max=p_max, imaginary=True, e_min=e_min,
                                   e_max=e_max, nb_unit_min=nb_unit_min,
                                   nb_unit_max=nb_unit_max,
                                   flow_direction='out',
                                   starting_cost=starting_cost,
                                   operating_cost=operating_cost,
                                   max_ramp_up=max_ramp_up,
                                   max_ramp_down=max_ramp_down,
                                   co2_out=co2_out, energy_type=energy_type,
                                   verbose=False,
                                   no_warn=no_warn)


class SquareProductionUnit(SquareEnergyUnit, VariableProductionUnit):
    """
    **Description**

        | Production unit with a fixed value and fixed duration.
        | Only the time of beginning can be modified.
        | Operation can be mandatory or not.

    **Attributes**

     * p : instantaneous power production (kW)
     * duration : duration of the power delivery (hours)
     * mandatory : indicates if the power delivery is mandatory or not
     * energy_type : type of energy ('Electrical', 'Heat', ...)

    """

    def __init__(self, time, name, p_square, duration, n_square,
                 t_between_sq, co2_out=None, starting_cost=None,
                 operating_cost=None, energy_type=None,
                 verbose=True, no_warn=True):
        duration /= time.DT
        if duration < 1:
            raise ValueError('The duration of operation of the '
                             'SquareProductionUnit should be longer than the '
                             'time step.')
        VariableProductionUnit.__init__(self, time=time, name=name,
                                        verbose=verbose)
        SquareEnergyUnit.__init__(self, time, name=name, p_square=p_square,
                                  n_square=n_square, t_square=duration,
                                  t_between_sq=t_between_sq,
                                  flow_direction='out',
                                  starting_cost=starting_cost,
                                  operating_cost=operating_cost,
                                  co2_out=co2_out, energy_type=energy_type,
                                  verbose=False,
                                  no_warn=no_warn)


class ShiftableProductionUnit(ShiftableEnergyUnit, VariableProductionUnit):
    """
    **Description**

        Production unit with shiftable production profile.

    **Attributes**

     * power_values : production profile to shift (kW)
     * mandatory : indicates if the production is mandatory : True
       or not : False
     * starting_cost : cost of the starting of the production
     * operating_cost : cost of the operation (€/kW)
     * energy_type : type of energy ('Electrical', 'Heat', ...)

    """

    def __init__(self, time, name: str, power_values, mandatory=True,
                 co2_out=None, starting_cost=None, operating_cost=None,
                 energy_type=None, verbose=True):

        VariableProductionUnit.__init__(self, time=time, name=name,
                                        verbose=verbose)
        ShiftableEnergyUnit.__init__(self, time, name=name,
                                     flow_direction='out',
                                     power_values=power_values,
                                     mandatory=mandatory, co2_out=co2_out,
                                     starting_cost=starting_cost,
                                     operating_cost=operating_cost,
                                     energy_type=energy_type,
                                     verbose=False)



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module defines the consumption units**

 The consumption_units module defines various classes of consumption units,
 from generic to specific ones.

 It includes :
    - ConsumptionUnit : simple consumption unit. It inherits from EnergyUnit,
      its power flow direction is always 'in'. \n
      3 Objectives are also available :

            * minimize consumption,
            * maximize consumption,
            * minimize consumption costs.

    - FixedConsumptionUnit :  consumption with a fixed load profile. It
      inherits from ConsumptionUnit.

    - VariableConsumptionUnit : consumption unit allowing for a variation of
      power between p_min et p_max. It inherits from ConsumptionUnit.

And also :
 - SeveralConsumptionUnit: Consumption unit based on a fixed consumption curve
   enabling to multiply several times (nb_unit) the same consumption profile
 - SeveralImaginaryConsumptionUnit: Consumption unit based on a fixed
   consumption curve enabling to multiply several times (nb_unit) the same
   consumption profile. Be careful, the solution may be imaginary as nb_unit
   can be continuous. The accurate number of the Consumption units should be
   calculated later
 - SquareConsumptionUnit: Consumption unit with a fixed value and fixed
   duration.
 - ShiftableConsumptionUnit: Consumption unit with shiftable consumption
   profile.

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from pulp import LpBinary
from .energy_units import *
del AssemblyUnit
from ...general.optimisation.elements import *
__docformat__ = "restructuredtext en"


class ConsumptionUnit(EnergyUnit):
    """
    **Description**

        Simple Consumption unit

    **Attributes**

         * p : instantaneous power demand (kW)
         * p_max : maximal instantaneous power demand (kW)
         * p_min : minimal instantaneous power demand (kW)
         * energy_type : type of energy ('Electrical', 'Heat', ...)
         * consumption_cost : cost associated to the energy consumption
     """

    def __init__(self, time, name, p=None, p_min=1e-5, p_max=1e+5, e_min=0,
                 e_max=1e6, co2_out=None, starting_cost=None,
                 consumption_cost=None, min_time_on=None, min_time_off=None,
                 max_ramp_up=None, max_ramp_down=None, availability_hours=None,
                 energy_type=None, verbose=True):
        EnergyUnit.__init__(self, time, name, flow_direction='in', p=p,
                            p_min=p_min, p_max=p_max, e_min=e_min, e_max=e_max,
                            co2_out=co2_out, starting_cost=starting_cost,
                            operating_cost=consumption_cost,
                            min_time_on=min_time_on,
                            min_time_off=min_time_off, max_ramp_up=max_ramp_up,
                            max_ramp_down=max_ramp_down,
                            availability_hours=availability_hours,
                            energy_type=energy_type,
                            verbose=verbose)

    # OBJECTIVES#
    def minimize_consumption(self, weight=1):
        """
        :param weight: Weight coefficient for the objective
        """
        self.minimize_energy(weight=weight)
        self.min_energy.name = 'min_consumption'

    def maximize_consumption(self, weight=1):
        """
        :param weight: Weight coefficient for the objective
        """
        self.minimize_energy(weight=-1 * weight)
        self.min_energy.name = 'max_consumption'

    def minimize_consumption_cost(self, weight=1):
        """

        :param weight: Weight coefficient for the objective
        """
        self.minimize_operating_cost(weight)
        self.min_operating_cost.name = 'min_consumption_cost'


class FixedConsumptionUnit(FixedEnergyUnit, ConsumptionUnit):
    """
    **Description**

        Consumption unit with a fixed consumption profile.

    **Attributes**

        * p : instantaneous power demand known in advance (kW)
        * energy_type : type of energy ('Electrical', 'Heat', ...)
        * consumption_cost : cost associated to the energy consumption

    """

    def __init__(self, time, name, p: list or dict or pd.DataFrame = None,
                 co2_out=None,
                 starting_cost=None, operating_cost=None, energy_type=None,
                 verbose=True):
        ConsumptionUnit.__init__(self, time=time, name=name, verbose=verbose)
        FixedEnergyUnit.__init__(self, time, name=name, p=p,
                                 flow_direction='in',
                                 starting_cost=starting_cost,
                                 operating_cost=operating_cost, co2_out=co2_out,
                                 energy_type=energy_type,
                                 verbose=False)


class VariableConsumptionUnit(VariableEnergyUnit, ConsumptionUnit):
    """
    **Description**

        Consumption unit with a variation of power between p_min et p_max.

    **Attributes**

        * p_max : maximal instantaneous power consumption (kW)
        * p_min : minimal instantaneous power consumption (kW)
        * energy_type : type of energy ('Electrical', 'Heat', ...)

    """

    def __init__(self, time, name, p_min=1e-5, p_max=1e+5, e_min=0,
                 e_max=1e6, co2_out=None, starting_cost=None,
                 operating_cost=None, min_time_on=None, min_time_off=None,
                 max_ramp_up=None, max_ramp_down=None, energy_type=None,
                 verbose=True, no_warn=True):
        ConsumptionUnit.__init__(self, time=time, name=name, verbose=verbose)
        VariableEnergyUnit.__init__(self, time, name=name,
                                    flow_direction='in', p_min=p_min,
                                    p_max=p_max, e_min=e_min, e_max=e_max,
                                    starting_cost=starting_cost,
                                    operating_cost=operating_cost,
                                    min_time_on=min_time_on,
                                    min_time_off=min_time_off,
                                    max_ramp_up=max_ramp_up,
                                    max_ramp_down=max_ramp_down,
                                    co2_out=co2_out, energy_type=energy_type,
                                    verbose=False, no_warn=no_warn)


class SeveralConsumptionUnit(VariableConsumptionUnit, SeveralEnergyUnit):
    """
    **Description**

        Consumption unit based on a fixed consumption curve enabling to multiply
        several times (nb_unit) the same consumption curve.

    **Attributes**

        * fixed_cons : fixed consumption curve

    """

    def __init__(self, time, name, fixed_cons, p_min=1e-5, p_max=1e+5, e_min=0,
                 e_max=1e6, nb_unit_min=0, nb_unit_max=None, co2_out=None,
                 starting_cost=None, operating_cost=None, max_ramp_up=None,
                 max_ramp_down=None, energy_type=None,
                 verbose=True, no_warn=True):
        VariableConsumptionUnit.__init__(self, time=time, name=name,
                                         verbose=verbose)
        SeveralEnergyUnit.__init__(self, time, name=name,
                                   fixed_power=fixed_cons, p_min=p_min,
                                   p_max=p_max, imaginary=False, e_min=e_min,
                                   e_max=e_max, nb_unit_min=nb_unit_min,
                                   nb_unit_max=nb_unit_max,
                                   flow_direction='in',
                                   starting_cost=starting_cost,
                                   operating_cost=operating_cost,
                                   max_ramp_up=max_ramp_up,
                                   max_ramp_down=max_ramp_down,
                                   co2_out=co2_out, energy_type=energy_type,
                                   verbose=False,
                                   no_warn=no_warn)


class SeveralImaginaryConsumptionUnit(VariableConsumptionUnit,
                                      SeveralEnergyUnit):
    """
    **Description**

        Consumption unit based on a fixed consumption curve enabling to multiply
        several times (nb_unit) the same consumption curve.
        Be careful, the solution may be imaginary as nb_unit can be
        continuous. The accurate number of the consumption unit should be
        calculated later.

    **Attributes**

        * fixed_cons : fixed consumption curve

    """

    def __init__(self, time, name, fixed_cons, p_min=1e-5, p_max=1e+5, e_min=0,
                 e_max=1e6, nb_unit_min=0, nb_unit_max=None, co2_out=None,
                 starting_cost=None, operating_cost=None, max_ramp_up=None,
                 max_ramp_down=None, energy_type=None,
                 verbose=True, no_warn=True):
        VariableConsumptionUnit.__init__(self, time=time, name=name,
                                         verbose=verbose)
        SeveralEnergyUnit.__init__(self, time, name=name,
                                   fixed_power=fixed_cons, p_min=p_min,
                                   p_max=p_max, imaginary=True, e_min=e_min,
                                   e_max=e_max, nb_unit_min=nb_unit_min,
                                   nb_unit_max=nb_unit_max,
                                   flow_direction='in',
                                   starting_cost=starting_cost,
                                   operating_cost=operating_cost,
                                   max_ramp_up=max_ramp_up,
                                   max_ramp_down=max_ramp_down,
                                   co2_out=co2_out, energy_type=energy_type,
                                   verbose=False,
                                   no_warn=no_warn)


class SquareConsumptionUnit(SquareEnergyUnit, VariableConsumptionUnit):
    """
    **Description**

        | Consumption unit with a fixed value and fixed duration.
        | Only the time of beginning can be modified
        | Operation can be mandatory or not

    **Attributes**

        * p : instantaneous power consumption (kW)
        * duration : duration of the power delivery (hours)
        * mandatory : indicates if the power delivery is mandatory or not
        * energy_type : type of energy ('Electrical', 'Heat', ...)
        * consumption_cost : cost associated to the energy consumption

    """

    def __init__(self, time, name, p_square, duration, n_square,
                 t_between_sq, co2_out=None, starting_cost=None,
                 operating_cost=None, energy_type=None,
                 verbose=True, no_warn=False):
        duration /= time.DT
        if duration < 1:
            raise ValueError('The duration of operation of the '
                             'SquareConsumptionUnit should be longer than the '
                             'time step.')
        duration = int(round(duration))
        VariableConsumptionUnit.__init__(self, time=time, name=name,
                                         verbose=verbose)
        SquareEnergyUnit.__init__(self, time, name=name, p_square=p_square,
                                  n_square=n_square, t_square=duration,
                                  t_between_sq=t_between_sq,
                                  flow_direction='in',
                                  starting_cost=starting_cost,
                                  operating_cost=operating_cost,
                                  co2_out=co2_out, energy_type=energy_type,
                                  verbose=False,
                                  no_warn=no_warn)


class ShiftableConsumptionUnit(ShiftableEnergyUnit, VariableConsumptionUnit):
    """
    **Description**

        Consumption unit with shiftable consumption profile.

    **Attributes**

        * power_values : consumption profile to shift (kW)
        * mandatory : indicates if the consumption is mandatory (True) or not
        (False)
        * starting_cost : cost of the starting of the consumption
        * operating_cost : cost of the operation (€/kW)
        * energy_type : type of energy ('Electrical', 'Heat', ...)

    """

    def __init__(self, time, name: str, power_values, mandatory=True,
                 co2_out=None, starting_cost=None, operating_cost=None,
                 energy_type=None, verbose=True):
        VariableConsumptionUnit.__init__(self, time=time, name=name,
                                         verbose=verbose)
        ShiftableEnergyUnit.__init__(self, time, name=name,
                                     flow_direction='in',
                                     power_values=power_values,
                                     mandatory=mandatory, co2_out=co2_out,
                                     starting_cost=starting_cost,
                                     operating_cost=operating_cost,
                                     energy_type=energy_type,
                                     verbose=False)



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module includes the optimization elements (quantities, constraints and
objectives) formulated in LP or MILP**

 - Quantity : related to the decision variable or parameter
 - Constraint : related to the optimization problem constraints
   4 categories of constraints :
    - Constraint: to define a constraint object
    - DefinitionConstraint: to calculate and define quantities or for physical
    constraints
    - TechnicalConstraint: linked with technical constraints
    - ActorConstraint: constraint decided by the actors
    2 types of constraints:
    - Constraints : basic constraint
    - DynamicConstraint: basic constraint depending on the time
 - Objective : related to the objective function

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import datetime
import numpy as np
import pandas as pd
from pulp import LpContinuous

__docformat__ = "restructuredtext en"


class Quantity:
    """
    **Description**

        Class that defines what is a quantity. A quantity can wether be a
        decision variable or a parameter, depending on the opt parameter

    **Attributes**

     - name (str) : the name of the quantity
     - description (str) :  a description of the meaning of the quantity
     - vtype (PuLP) : the variable type, depending on PuLP

        * LpBinary (binary variable)
        * LpInteger (integer variable)
        * LpContinuous (continuous variable)
     - vlen (int) : size of the variable
     - unit (str) : unit of the quantity
     - opt (binary)

        * True: this is an optimization variable
        * False: this is a constant - a parameter
     - value (float, list, dict) : value (unused if opt=True)
     - ub, lb : upper and lower bounds
     - parent (OptObject) : the quantity belongs to this unit

    .. note::
        Make sure that all the modifications on Quantity are made before
        adding the unit to the Model
    """

    def __init__(self, name="var0", opt=True, unit="s.u", vlen=None,
                 value=None, description="", vtype=LpContinuous,
                 lb=None, ub=None, parent=None):
        """

        :param name: (str) name of the quantity
        :param opt: (binary) optimisation parameter
        :param unit: (str) unit of the quantity
        :param vlen: (int) size of the quantity value
        :param value: (float, list or dict) value of the quantity
        :param description: (str) description of the quantity
        :param vtype: (type) the variable type according to PuLP
        :param lb: (float or int) : lower bound of the value of the quantity
        :param ub: (float or int) : upper bound of the value of the quantity
        :param parent: (OptObject) parent unit
        """

        self.name = name
        if vlen is None:
            try:
                vlen = len(value)
            except TypeError:
                vlen = 1

        if parent is None:
            print('Warning : Parent argument of {0} is None'.format(self.name))

        self.vlen = vlen
        self.description = description
        self.unit = unit
        self.lb = lb
        self.ub = ub
        self.vtype = vtype
        self.value = {}

        if not isinstance(vlen, int):
            raise TypeError('vlen should be of type int !')

        if isinstance(opt, bool):
            # The optimisation parameter is a boolean
            if isinstance(value, (int, float)):
                # The value is an int or a float
                self.opt = opt
                if value:
                    # If the value is not None, the opt parameter should be
                    # False
                    self.opt = False
                if vlen <= 1:
                    self.value = value
                else:
                    # vlen should be 1 for ints and floats values
                    raise ValueError(
                        'value\'s size is given as vlen = {0}, but the value '
                        'is actually a {1}'.format(vlen, type(value)))

            elif isinstance(value, list):
                # The value is a list
                if value:
                    # If value is not None, the opt parameter is a list of
                    # False
                    self.opt = []
                    for i in range(vlen):
                        self.opt.append(False)
                if len(value) == vlen:
                    self.value = value
                else:
                    # vlen should be the list length
                    raise ValueError(
                        'value\'s size ({0}) should be vlen ({1})'.format(
                            len(value), vlen))

            elif isinstance(value, dict):
                # The value is a dict
                if any(isinstance(n, (int, float)) for n in list(
                        value.values())):
                    # At least one value of the dict is specified
                    self.opt = {}
                    # The opt parameter is a dict of False
                    self.opt.update({i: False for i in list(value.keys())})
                    print('At least one value is specified in the dict '
                          'value : the opt parameter is set as a dict of '
                          'False with the value keys')
                else:
                    self.opt = {}
                    self.opt.update({i: opt for i in list(value.keys())})

                if len(value) == vlen:
                    self.value = value
                else:
                    # vlen should be the dict value size
                    raise ValueError(
                        'value\'s size ({0}) should be vlen ({1})'.format(
                            len(value), vlen))

            elif value is None:
                if not opt:
                    # The quantity should have a specified value if
                    # opt=False OR opt should be set to True if the Quantity
                    #  is a decision variable
                    raise Exception('If the Quantity.opt == False, '
                                    'Quantity.value cannot be None. '
                                    'Value is required, or opt must be set '
                                    'to True !')
                elif vlen >= 2:
                    # if opt=True, the value is given the default value 0,
                    # or a list of 0 if vlen>=2
                    self.opt = {i: opt for i in range(vlen)}
                    self.value = {i: 0 for i in range(vlen)}

                else:
                    self.opt = opt
                    self.value = 0

            else:
                # Value should be either an int, a float, a list, a dict or
                # None
                raise TypeError('Unknown type for argument \'value\'')

        elif isinstance(opt, (list, dict)):
            raise Exception('Not implemented yet...')

        else:
            # opt should be a boolean, a list or a dict
            raise TypeError('Unknown type for argument \'opt\'')

        self.parent = parent
        self._add_quantity_list()

    def __str__(self):
        # String special method
        return str(self.value)

    def __repr__(self):
        # Repr special method
        if isinstance(self.opt, bool):
            exp = '<OMEGALPES.general.optimisation.elements.Quantity : (' \
                  'name:\'{0}\', opt:{1}, vlen:{1})> \n'.format(self.name,
                                                                self.opt,
                                                                self.vlen)
        elif isinstance(self.opt, list):
            exp = '<OMEGALPES.general.optimisation.elements.Quantity : (' \
                  'name:\'{0}\', opt[0]:{1}, vlen:{2})> \n'.format(self.name,
                                                                   self.opt[0],
                                                                   self.vlen)

        elif isinstance(self.opt, dict):
            exp = '<OMEGALPES.general.optimisation.elements.Quantity : (' \
                  'name:\'{0}\', opt:{1}, vlen:{2})> \n'.format(self.name,
                                                                list(
                                                                    self.opt.values()),
                                                                self.vlen)
        return exp

    def __float__(self):
        # Special method float (only works for int and float values,
        # not list and dict ones)
        return float(self.value)

    def get_value(self):
        """ return the value of the quantity according the type of the value
        in order to be able to use it easily in print and plot methods:
        int -> int
        float -> float
        list -> list
        dict -> list
        ndarray --> list
        """
        if isinstance(self, Quantity):
            if isinstance(self.value, dict):
                return list(self.value.values())

            if isinstance(self.value, np.ndarray):
                return list(self.value.values().tolist())

            else:
                return self.value
        else:
            raise TypeError('you can only use the get_value() method on'
                            'quantities')

    def get_value_with_date(self, unit=None):
        """ return the values of the quantity associated to a date in a
        dataframe if the values are a list or a dict (only).
        """
        if not unit:
            unit = self.unit
        else:
            pass

        if isinstance(self, Quantity):
            if isinstance(self.value, dict):

                df = pd.DataFrame(data=list(self.value.values()),
                                  index=self.parent.time.DATES,
                                  columns=['Value in {}'.format(unit)])
                return df

            if isinstance(self.value, list):
                df = pd.DataFrame(data=list(self.value),
                                  index=self.parent.time.DATES,
                                  columns=['Value in {}'.format(unit)])
                return df

            else:
                raise TypeError(
                    'the value {} should be a list or a dict'.format(
                        self.name))

        else:
            raise TypeError('you can only use the get_value() method on'
                            'quantities')

    def _add_quantity_list(self):
        if self.parent:
            if self not in self.parent._quantities_list:
                self.parent._quantities_list.append(self)
        else:
            Warning("the quantity {} should have a parent otherwise "
                    "the quantity won't be added to the parent "
                    "_quantities_list".format(self.name))


class Constraint:
    """
    **Description**

        Class that defines a constraint object

    **Attributes**

     - name: name of the constraint
     - description: a description of the constraint
     - active:
       False = non-active constraint;
       True = active constraint
     - exp: (str) : expression of the constraint
     - parent: (unit) : this constraint belongs to this unit

    .. note::
        Make sure that all the modifications on Constraints are made before
        adding the unit to the Model (OptimisationModel.addUnit()).

    """

    def __init__(self, exp, name='CST0', description='', active=True,
                 parent=None):
        """

        :param exp: (str) expression of the constraint
        :param name: (str) name of the constraint
        :param description: (str) description of the constrain
        :param active: (bool) is the constraint active or not
        :param parent: (OptObject) parent unit
        """
        self.name = name
        self.description = description
        self.exp = exp
        self.active = active
        self.parent = parent
        self._add_cst_list()

    def _add_cst_list(self):
        if self.parent:
            if self not in self.parent._constraints_list:
                self.parent._constraints_list.append(self)
        else:
            Warning("the constraint {} should have a parent otherwise "
                    "the constraint won't be added to the parent "
                    "_constraints_list".format(self.name))

    def _deactivate_constraint(self):
        """ An constraint can be deactivated in order to debug (only!)
        """
        self.active = False


class DynamicConstraint(Constraint):
    """
    **Description**

        Class that defines a constraint depending on the time.
        NB : Mandatory for PuLP

    """

    def __init__(self, exp_t, t_range='for t in time.I', name='DCST0',
                 description='dynamic constraint', active=True, parent=None):
        """

        :param exp_t (str): constraint expression time-dependant
        :param t_range (str): time range defining when the constraint is set
        :param name (str): name of the constraint
        :param description (str): description of the constraint
        :param active (bool): is the constraint activated or deactivated
        :param parent: unit containing the constraint
        """
        if t_range[0] == ' ':
            # If the t_range starts with an empty space, it is deleted
            t_range = t_range[1:]
        exp = exp_t + ' ' + t_range

        Constraint.__init__(self, exp, name, description, active, parent)

        self.exp_t = exp_t
        self.t_range = t_range


class DefinitionConstraint(Constraint):
    """
    **Description**

        Defining a category of constraint: DefinitionConstraint for imposed
        mathematical relation constraints, which may be physical. They are
        used to calculate and define quantities, or to reprensent physical
        phenomenon.
        Must be different from Technical and Actor constraints.
        This kind of constraint is by nature non-negotiable

        Definition constraint: added to the model to calculate a quantity or
        define it considering an other.
            example: the definition of e_tot: calc_e_tot (e_tot = LpSum(e))
        Physical constraint: to model a constraint which is linked to the
        physics.
            example: not implemented yet, see DefinitionDynamicConstraint
    """

    def __init__(self, exp, name='DefCST0', description='Constraint for '
                                                        'definitions',
                 active=True,
                 parent=None):
        Constraint.__init__(self, exp=exp, name=name,
                            description=description, active=active,
                            parent=parent)


class DefinitionDynamicConstraint(DynamicConstraint, DefinitionConstraint):
    """
    **Description**

        Defining a category of constraint: DefinitionDynamicConstraint
        on the time.  (see: DynamicConstraint, DefinitionConstraint)
        Must be different from Technical and Actor constraints.
        This kind of constraint is by nature non-negotiable

        DefinitionDynamicConstraint: to calculate a quantity
        or define it considering an other quantity and depending on the time.
            example: the definition of a capacity def_capacity
            (energy[t] <= capacity at each time step t)
        Physical constraint: to model physical constraints.
            example: the power balance in an Energy node: power_balance
           (LpSum(p_production_unit[t] for the set of production units
           connected to the energy node) = LpSum(p_consumption_unit[t] for the
           consumption units connected to the energy node at each time step t)
    """

    def __init__(self, exp_t, t_range='for t in time.I', name='DefDynCST0',
                 description='dynamic constraint for definition',
                 active=True, parent=None):
        DefinitionConstraint.__init__(self, exp='', name=name,
                                      description=description, active=active,
                                      parent=parent)
        DynamicConstraint.__init__(self, exp_t=exp_t,
                                   t_range=t_range, name=name,
                                   description=description,
                                   active=active, parent=parent)


class TechnicalConstraint(Constraint):
    """
    **Description**

        Defining a category of constraint: TechnicalConstraint.
        To model a constraint which is linked to
        technical issues
        Must be different from Definition and Actor constraints.
        This kind of constraint is mainly a negotiable constraint.


            example: not implemented yet, see TechnicalDynamicConstraints

    """

    def __init__(self, exp, name='TechCST0', description='Technical '
                                                         'constraint',
                 active=True,
                 parent=None):
        Constraint.__init__(self, exp, name, description, active, parent)
        self._add_tech_cst_list()

    def _add_tech_cst_list(self):
        if self.parent:
            if self not in self.parent._technical_constraints_list:
                self.parent._technical_constraints_list.append(self)
            else:
                Warning("the technical constraint {} should have a parent "
                        "otherwise the constraint won't be added to the "
                        "parent "
                        "_technical_constraints_list".format(self.name))

    def deactivate_constraint(self):
        """ An technical constraint can be deactivated :
             - To compare scenarios
             - To try a less constrained problem
        """
        self.active = False


class TechnicalDynamicConstraint(DynamicConstraint, TechnicalConstraint):
    """
    **Description**

        Defining a category of constraint: TechnicalDynamicConstraint
        depending on the time. (see: DynamicConstraint, TechnicalConstraint)
        Must be different from Definition and Actor constraints.
        This kind of constraint is mainly a negotiable constraint.

            example: an energy unit may not be able to stop in one time
            step, but several. The power decreases as a ramp shape:
            set_max_ramp_down

    """

    def __init__(self, exp_t, t_range='for t in time.I', name='TechCST0',
                 active=True,
                 description='Technical and dynamic constraint',
                 parent=None):
        TechnicalConstraint.__init__(self, exp='', name=name,
                                     description=description,
                                     active=active, parent=parent)
        DynamicConstraint.__init__(self, exp_t=exp_t, t_range=t_range,
                                   name=name,
                                   description=description, active=active,
                                   parent=parent)


class ActorConstraint(Constraint):
    """
    **Description**

        Defining a category of constraint: ActorConstraint
        (see: DynamicConstraint, TechnicalConstraint)
        To model a constraint which is due to actor decisions.
        Must be different from Definition and Technical constraints.
        This kind of constraint is mainly a negotiable constraint.

            example: to have a minimal level of energy consumption
            over a period. Constraint : min_energy

    """

    def __init__(self, exp, name='ActCST0', description='', active=True,
                 parent=None):
        Constraint.__init__(self, exp, name, description, active, parent)
        self._add_actor_cst_list()

    def _add_actor_cst_list(self):
        if self.parent:
            if self not in self.parent._actor_constraints_list:
                self.parent._actor_constraints_list.append(self)
            else:
                Warning("the actor's constraint {} should have a parent "
                        "otherwise the constraint won't be added to the "
                        "parent "
                        "_actor_constraints_list".format(self.name))

    def deactivate_constraint(self):
        """ An actor's constraint can be deactivated :
             - To compare scenarios
             - To try a less constrained problem
        """
        self.active = False


class ActorDynamicConstraint(DynamicConstraint, ActorConstraint):
    """
    **Description**

        Defining a category of constraint: ActorDynamicConstraint
        (see: DynamicConstraint, ActorConstraint)
        Must be different from Definition and Technical constraints.
        This kind of constraint is mainly a negotiable constraint.

            example: to operate an energy unit only on defined
            periods. Constraint: daily_dynamic_constraint

    """

    def __init__(self, exp_t, t_range='for t in time.I', name='ActDynCST0',
                 active=True,
                 description='Actor and dynamic constraint',
                 parent=None):
        ActorConstraint.__init__(self, exp='', name=name,
                                 description=description,
                                 active=active, parent=parent)
        DynamicConstraint.__init__(self, exp_t=exp_t, t_range=t_range,
                                   name=name,
                                   description=description, active=active,
                                   parent=parent)


class ExternalConstraint(Constraint):
    """
    **Description**

        +++++++++++++++++++++++++++++++++++++++++++++
          **DEPRECATED: please use ActorConstraint**
        +++++++++++++++++++++++++++++++++++++++++++++

        Defining a special type of constraint: the external constraint
         - This constraint does not translate a physical constraint
         - This constraint defines an external constraint, which could be
           relaxed
 """

    def __init__(self, exp, name='ExCST0', description='', active=True,
                 parent=None):
        Constraint.__init__(self, exp, name, description, active, parent)

        import warnings
        deprecated_msg = "DEPRECATED: HourlyDynamicConstraint is deprecated " \
                         "and will be removed in omegalpes version 0.4 " \
                         "Please use DailyDynamicConstraint that allows to " \
                         "use hour and minutes in the daily constraint " \
                         "definition"
        warnings.warn(deprecated_msg, DeprecationWarning)

        self._add_ext_cst_list()

    def _add_ext_cst_list(self):
        if self.parent:
            if self not in self.parent._external_constraints_list:
                self.parent._external_constraints_list.append(self)
            else:
                Warning("the external constraint {} should have a parent "
                        "otherwise the constraint won't be added to the "
                        "parent "
                        "_external_constraints_list".format(self.name))

    def deactivate_constraint(self):
        """ An external constraint can be deactivated :
             - To compare scenarios
             - To try a less constrained problem
        """
        self.active = False


class ExtDynConstraint(DynamicConstraint, ExternalConstraint):
    """
    **Description**

        +++++++++++++++++++++++++++++++++++++++++++++++++++
          **DEPRECATED: please use ActorDynamicConstraint**
        ++++++++++++++++++++++++++++++++++++++++++++++++++++

        Defining a constraint both external and dynamic
        (see: DynamicConstraint, ExternalConstraint)
    """

    def __init__(self, exp_t, t_range='for t in time.I', name='EDCST0',
                 active=True,
                 description='Non-physical and dynamic constraint',
                 parent=None):
        import warnings
        deprecated_msg = "DEPRECATED: HourlyDynamicConstraint is deprecated " \
                         "and will be removed in omegalpes version 0.4 " \
                         "Please use DailyDynamicConstraint that allows to " \
                         "use hour and minutes in the daily constraint " \
                         "definition"
        warnings.warn(deprecated_msg, DeprecationWarning)

        ExternalConstraint.__init__(self, exp='', name=name,
                                    description=description,
                                    active=active, parent=parent)
        DynamicConstraint.__init__(self, exp_t=exp_t, t_range=t_range,
                                   name=name,
                                   description=description, active=active,
                                   parent=parent)


class HourlyDynamicConstraint(ActorDynamicConstraint):
    """
    **Description**

        +++++++++++++++++++++++++++++++++++++++++++++++++
          **DEPRECATED: please use DailyDynamicConstraint**
        +++++++++++++++++++++++++++++++++++++++++++++++++

        Class that defines an dynamic contraint for a time range

        Ex: Constraint applying between 7am and 10pm

        ex_cst = HourlyDynamicConstraint(exp_t, time, init_h=7, final_h=22,
        name='ex_cst')

    **Attributes**

     - name (str) : name of the constraint
     - exp_t (str) : expression of the constraint
     - init_h (int) : hour of beginning of the constraint [0-23]
     - final_h (int) : hour of end of the constraint [1-24]
     - description (str) : description of the constraint
     - active (bool) : defines if the constraint is active or not
     - parent (OptObject) : parent of the constraint
    """

    def __init__(self, exp_t, time, init_h: int = 0, final_h: int = 24,
                 name='HDCST0', description='hourly dynamic constraint',
                 active=True, parent=None):

        import warnings
        deprecated_msg = "DEPRECATED: HourlyDynamicConstraint is deprecated " \
                         "and will be removed in omegalpes version 0.4 " \
                         "Please use DailyDynamicConstraint that allows to " \
                         "use hour and minutes in the daily constraint " \
                         "definition"
        warnings.warn(deprecated_msg, DeprecationWarning)

        if init_h > final_h:
            raise ValueError('final_h {0} should be greater than init_h '
                             '{1}'.format(final_h, init_h))
        if final_h > 24:
            raise ValueError('final_h {} should be lower than '
                             '24'.format(final_h))

        index_list = []  # Initializing the list of index

        # For all days in the time periods, select the hour range
        for day in time.get_days:
            starting_date = datetime.datetime(year=day.year, month=day.month,
                                              day=day.day, hour=init_h)

            if final_h == 24:
                end = day + datetime.timedelta(days=1)

            else:
                end = datetime.datetime(year=day.year, month=day.month,
                                        day=day.day, hour=final_h)

            try:
                index_date = time.get_index_for_date_range(
                    starting_date=starting_date, end=end)
                index_list += index_date

            except ValueError:
                pass  # Exception when no values for the last day for instance

        t_range = 'for t in {}'.format(index_list)

        ActorDynamicConstraint.__init__(self, exp_t, t_range=t_range,
                                        name=name,
                                        description=description, active=active,
                                        parent=parent)


class DailyDynamicConstraint(ActorDynamicConstraint):
    """
    **Description**

        Class that defines a daily dynamic constraint for a time range

        Ex: Constraint applying between 7am and 10:30pm

        ex_cst = DailyDynamicConstraint(exp_t, time, init_h="7:00",
        final_h="10:30", name='ex_cst')

    **Attributes**

     - name (str) : name of the constraint
     - exp_t (str) : expression of the constraint
     - init_time (str) : starting time of the constraint in the format:
     "HH:MM", consistent with the dt value.
     - final_time (str) : ending time of the constraint in the format:
     "HH:MM", consistent with the dt value.
     - description (str) : description of the constraint
     - active (bool) : defines if the constraint is active or not
     - parent (OptObject) : parent of the constraint
    """

    def __init__(self, exp_t, time, init_time: str = "00:00", final_time: str =
    "23:00", name='daily_dynamic_constraint',
                 description='daily dynamic constraint',
                 active=True, parent=None):

        (init_h_str, init_min_str) = init_time.split(":")
        (final_h_str, final_min_str) = final_time.split(":")

        init_h = int(init_h_str)
        init_min = int(init_min_str)
        final_h = int(final_h_str)
        final_min = int(final_min_str)

        if final_h != 00 or final_min != 00:
            if init_h > final_h or (init_h == final_h and init_min >
                                    final_min):
                raise ValueError('The final time {0} should be greater than '
                                 'init_time {1}'.format(final_time, init_time))

        if final_h >= 24:
            raise ValueError('The final hour {} should be lower than '
                             '24'.format(final_h))
        if init_min > 59 or final_min > 59:
            raise ValueError('The defined initial and final minutes should '
                             'have values lower than 60 !')

        index_list = []  # Initializing the list of index
        # For all days in the time periods, select the hour range
        for day in time.get_days:
            starting_date = datetime.datetime(year=day.year, month=day.month,
                                              day=day.day, hour=init_h,
                                              minute=init_min)

            if final_h == 00 and final_min == 00:
                end_date = day + datetime.timedelta(days=1)
            else:
                end_date = datetime.datetime(year=day.year, month=day.month,
                                             day=day.day, hour=final_h,
                                             minute=final_min)

            try:
                index_date = time.get_index_for_date_range(
                    starting_date=starting_date, end=end_date)
                index_list += index_date

            except ValueError:
                # Exception when no values for the last day for instance
                end_date = time.get_date_for_index(time.LEN - 1)
                index_date = time.get_index_for_date_range(
                    starting_date=starting_date, end=end_date)
                index_list += index_date

        t_range = 'for t in {}'.format(index_list)

        ActorDynamicConstraint.__init__(self, exp_t, t_range=t_range,
                                        name=name,
                                        description=description, active=active,
                                        parent=parent)


class Objective:
    """
     **Description**

        Class that defines an optimisation objective

     **Attributes**

      - name (str) :
      - description (str) :
      - active (bool) :
      - exp (str) :
      - weight (float) : weighted factor of the objective
      - parent (unit)
      - unit (str) : unit of the cost expression

     .. note::
        Make sure that all the modifications on Objectives are made before
        adding the unit to the OptimisationModel, otherwise, it won't be taken
        into account

     """

    def __init__(self, exp, name='OBJ0', description='', active=True, weight=1,
                 unit='s.u.', parent=None):
        self.name = name
        self.exp = exp
        self.description = description
        self.active = active
        self.weight = weight
        self.parent = parent
        self.unit = unit
        self._add_objective_list()

    def _add_objective_list(self):
        if self.parent:
            if self not in self.parent._objectives_list:
                self.parent._objectives_list.append(self)
        else:
            Warning("the objective {} should have a parent otherwise "
                    "the objective won't be added to the parent "
                    "_objectives_list".format(self.name))



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module describes the Prosumer (producer and consumer) actor**

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from omegalpes.actor.operator_actors.consumer_actors import Consumer
from omegalpes.actor.operator_actors.producer_actors import Producer
from omegalpes.general.optimisation.elements import *
from omegalpes.general.utils.maths import def_abs_value

__docformat__ = "restructuredtext en"


class Prosumer(Consumer, Producer):
    """
    **Description**

        Prosumer class inherits from the the class OperatorActor, Consumer
        and Producer. It enables one to model an actor which is at the same
        time an energy producer and consumer
    """

    def __init__(self, name, operated_consumption_unit_list,
                 operated_production_unit_list, operated_node_list=None,
                 verbose=True):
        Consumer.__init__(self, name=name,
                          operated_unit_list=operated_consumption_unit_list,
                          operated_node_list=operated_node_list,
                          verbose=verbose)
        Producer.__init__(self, name=name,
                          operated_unit_list=operated_production_unit_list,
                          operated_node_list=operated_node_list,
                          verbose=False)
        self.operated_consumption_unit_list = operated_consumption_unit_list
        self.operated_production_unit_list = operated_production_unit_list
        self.operated_unit_list = operated_consumption_unit_list + \
                                  operated_production_unit_list

    def _check_operated_consumption_list(self, obj_operated_unit_list):
        """
        Check if the objective must be applied to the whole energy units
        under the scope of responsibility (if obj_operated_unit_list is
        empty); or to selected units and in this case, if the units are under
        the scope of responsibility. Only for consumption units

        :param obj_operated_unit_list: List of units on which the
            objective will be applied. Might be empty.
        :return: the list of the operated energy unit in order to apply
            the objective on these units
        """
        final_operated_unit_list = []
        if not obj_operated_unit_list:
            final_operated_unit_list = self.operated_consumption_unit_list
        else:
            final_operated_unit_list = self._check_operated_unit_list(
                obj_operated_unit_list)
        return final_operated_unit_list

    def _check_operated_production_list(self, obj_operated_unit_list):
        """
        Check if the objective must be applied to the whole energy units
        under the scope of responsibility (if obj_operated_unit_list is
        empty); or to selected units and in this case, if the units are under
        the scope of responsibility. Only for production units

        :param obj_operated_unit_list: List of units on which the
            objective will be applied. Might be empty.
        :return: the list of the operated energy unit in order to apply
            the objective on these units
        """
        final_operated_unit_list = []
        if not obj_operated_unit_list:
            final_operated_unit_list = self.operated_production_unit_list
        else:
            final_operated_unit_list = self._check_operated_unit_list(
                obj_operated_unit_list)
        return final_operated_unit_list

    # OBJECTIVE #
    def maximize_conso_prod_match(self,
                                  time,
                                  obj_operated_consumption_unit_list=None,
                                  obj_operated_production_unit_list=None,
                                  weight=1):
        """
        To create the objective in order to match at each time the consumption
        with the local production of the prosumer's units (all or part of them).

        :param obj_operated_consumption_unit_list: List of consumption units on
            which the objective will be applied. Might be empty.
        :param obj_operated_production_unit_list: List of production units on
            which the objective will be applied. Might be empty.
        :param weight: Weight coefficient for the objective
        """
        final_operated_consumption_unit_list = \
            self._check_operated_consumption_list(
                obj_operated_consumption_unit_list)
        final_operated_production_unit_list = \
            self._check_operated_production_list(
                obj_operated_production_unit_list)

        consumption_p_string = ''
        production_p_string = ''
        for consumption_unit in final_operated_consumption_unit_list[:-1]:
            consumption_p_string += '{}_p[t] + '.format(consumption_unit.name)
        consumption_p_string += '{}_p[t]'.format(
            final_operated_consumption_unit_list[-1].name)
        for production_unit in final_operated_production_unit_list[:-1]:
            production_p_string += '{}_p[t] + '.format(production_unit.name)
        production_p_string += '{}_p[t]'.format(
            final_operated_production_unit_list[-1].name)

        self.conso_prod_match = Quantity(name='conso_prod_match',
                                         description='conso_prod_match '
                                                     'for objective',
                                         vlen=time.LEN, parent=self)

        self.calc_conso_prod_match = DefinitionDynamicConstraint(
            name='calc_conso_prod_match',
            exp_t='{0}_conso_prod_match[t] == ({1} - ( {2} )) * time.DT'.format(
                self.name, production_p_string, consumption_p_string),
            t_range='for t in time.I', parent=self)

        self.conso_prod_match_abs = def_abs_value(
            self.conso_prod_match, q_min=-1e+5, q_max=1e+5)

        self.max_conso_prod_match_pos = Objective(
            name='max_conso_prod_match_pos',
            exp='lpSum({}_conso_prod_match_pos[t] for t in time.I)'
                .format(self.name),
            weight=weight, parent=self)

        self.max_conso_prod_match_neg = Objective(
            name='max_conso_prod_match_neg',
            exp='lpSum({}_conso_prod_match_neg[t] for t in time.I)'
                .format(self.name),
            weight=weight, parent=self)

    def maximize_selfconsumption_rate(self, time,
                                      obj_operated_production_unit_list=None,
                                      obj_operated_consumption_unit_list=None,
                                      obj_operated_selfconsummed_production_export_list=None,
                                      obj_operated_selfconsummed_production_unit_list=None,
                                      weight=1):
        """
        To create the objective in order to maximize the selfconsumption rate
        of the prosumer's units (all or part of them) WHILE maximizing the
        load matching selfconsummed production is calculated with the export
        nodes.

        Selfconsumption rate = selfconsummed production / total production

        :param obj_operated_production_unit_list: List of production units on
            which the objective will be applied. Might be empty.
        :param obj_operated_selfconsummed_production_export_list: List of
            production exports from nodes on which the objective will be
            applied. Might be empty.
        :param obj_operated_selfconsummed_production_unit_list: List of
            production units on which the objective will be applied. Might be
            empty.
        :param weight: Weight coefficient for the objective
        """
        self.maximize_conso_prod_match(time,
                                       obj_operated_consumption_unit_list=
                                       obj_operated_consumption_unit_list,
                                       obj_operated_production_unit_list=
                                       obj_operated_production_unit_list)

        final_operated_production_unit_list = \
            self._check_operated_production_list(
                obj_operated_production_unit_list)
        if obj_operated_selfconsummed_production_export_list:
            final_operated_selfconsummed_production_export_list = \
                self._check_operated_unit_list(
                    obj_operated_selfconsummed_production_export_list)
        if obj_operated_selfconsummed_production_unit_list:
            final_operated_selfconsummed_production_unit_list = \
                self._check_operated_unit_list(
                    obj_operated_selfconsummed_production_unit_list)

        production_p_string = ''
        sc_production_unit_p_string = ''
        sc_production_export_p_string = ''
        for production_unit in final_operated_production_unit_list[:-1]:
            production_p_string += '{}_p[t] + '.format(production_unit.name)
        production_p_string += '{}_p[t]'.format(
            final_operated_production_unit_list[-1].name)
        if obj_operated_selfconsummed_production_export_list:
            for sc_production_export in \
                    final_operated_selfconsummed_production_export_list[:-1]:
                sc_production_export_p_string += '{0}_{1}[t] + '.format(
                    sc_production_export.parent.name, sc_production_export.name)
            sc_production_export_p_string += '{0}_{1}[t]'.format(
                final_operated_selfconsummed_production_export_list[
                    -1].parent.name,
                final_operated_selfconsummed_production_export_list[-1].name)
        if obj_operated_selfconsummed_production_unit_list:
            for sc_production_unit in \
                    final_operated_selfconsummed_production_unit_list[:-1]:
                sc_production_unit_p_string += '{}_p[t] + '.format(
                    sc_production_unit.name)
            sc_production_unit_p_string += '{}_p[t]'.format(
                final_operated_selfconsummed_production_unit_list[-1].name)

        self.selfconsumption_rate = Quantity(
            name='selfconsumption_rate', description='maximize selfconsumption'
                                                     'rate WHILE maximizing the'
                                                     'load matching',
            vlen=time.LEN, parent=self)

        self.calc_selfconsumption_rate = DefinitionDynamicConstraint(
            name='calc_selfconsumption_rate',
            exp_t='{0}_selfconsumption_rate[t] == ({1} + {2} - ( {3} '
                  ')) * time.DT'.format(self.name,
                                        sc_production_export_p_string,
                                        sc_production_unit_p_string, production_p_string),
            t_range='for t in time.I', parent=self)

        self.selfconsumption_rate_abs = def_abs_value(
            self.selfconsumption_rate, q_min=-1e+5, q_max=1e+5)

        self.max_selfconsumption_rate_pos = Objective(
            name='max_selfconsumption_rate_pos',
            exp='lpSum({}_selfconsumption_rate_pos[t] for t in time.I)'
                .format(self.name),
            weight=weight, parent=self)

        self.max_selfconsumption_rate_neg = Objective(
            name='max_selfconsumption_rate_neg',
            exp='lpSum({}_selfconsumption_rate_neg[t] for t in time.I)'
                .format(self.name),
            weight=weight, parent=self)

    def maximize_selfproduction_rate(self, time,
                                     obj_operated_consumption_unit_list=None,
                                     obj_operated_production_unit_list=None,
                                     obj_operated_selfproduced_consumption_export_list=None,
                                     obj_operated_selfproduced_consumption_unit_list=None,
                                     weight=1):
        """
        To create the objective in order to maximize the selfproduction rate
        of the prosumer's units (all or part of them) WHILE maximizing the
        load matching selfproduced consumption may required export nodes

        Selfproduction rate = selfproduced consumption / total consumption

        :param obj_operated_consumption_unit_list: List of consumption units on
            which the objective will be applied. Might be empty.
        :param obj_operated_selfproduced_consumption_export_list: List of
            production exports from nodes on which the objective will be
            applied.
            Might be empty.
        :param obj_operated_selfproduced_consumption_unit_list: List of
            production units on which the objective will be applied.
            Might be empty.
        :param weight: Weight coefficient for the objective
        """
        self.maximize_conso_prod_match(time,
                                       obj_operated_consumption_unit_list=
                                       obj_operated_consumption_unit_list,
                                       obj_operated_production_unit_list=
                                       obj_operated_production_unit_list)

        final_operated_consumption_unit_list = \
            self._check_operated_consumption_list(
                obj_operated_consumption_unit_list)
        if obj_operated_selfproduced_consumption_export_list:
            final_operated_sp_consumption_export_list = \
                self._check_operated_unit_list(
                    obj_operated_selfproduced_consumption_export_list)
        if obj_operated_selfproduced_consumption_unit_list:
            final_operated_sp_consumption_unit_list = \
                self._check_operated_unit_list(
                    obj_operated_selfproduced_consumption_unit_list)

        consumption_p_string = ''
        sp_consumption_export_p_string = ''
        sp_consumption_unit_p_string = ''
        for consumption_unit in final_operated_consumption_unit_list[:-1]:
            consumption_p_string += '{}_p[t] + '.format(consumption_unit.name)
        consumption_p_string += '{}_p[t]'.format(
            final_operated_consumption_unit_list[-1].name)
        if obj_operated_selfproduced_consumption_export_list:
            for sp_consumption_export in \
                    final_operated_sp_consumption_export_list[:-1]:
                sp_consumption_export_p_string += '{0}_{1}[t] + '.format(
                    sp_consumption_export.parent.name,
                    sp_consumption_export.name)
            sp_consumption_export_p_string += '{0}_{1}[t]'.format(
                final_operated_sp_consumption_export_list[-1].parent.name,
                final_operated_sp_consumption_export_list[
                    -1].name)
        if obj_operated_selfproduced_consumption_unit_list:
            for sp_consumption_unit in \
                    final_operated_sp_consumption_unit_list[:-1]:
                sp_consumption_unit_p_string += '{}_p[t] + '.format(
                    sp_consumption_unit.name)
                sp_consumption_unit_p_string += '{}_p[t]'.format(
                    final_operated_sp_consumption_unit_list[-1].name)

        self.selfproduction_rate = Quantity(
            name='selfproduction_rate', description='maximize selfconsumption'
                                                    'rate WHILE maximizing the'
                                                    'load matching',
            vlen=time.LEN, parent=self)

        self.calc_selfproduction_rate = DefinitionDynamicConstraint(
            name='calc_selfproduction_rate',
            exp_t='{0}_selfproduction_rate[t] == ({1} + {2} - ( {3} '
                  ')) * time.DT'.format(
                self.name, sp_consumption_export_p_string,
                sp_consumption_unit_p_string, consumption_p_string),
            t_range='for t in time.I', parent=self)

        self.selfproduction_rate_abs = def_abs_value(
            self.selfproduction_rate, q_min=-1e+5, q_max=1e+5)

        self.max_selfproduction_rate_pos = Objective(
            name='max_selfproduction_rate_pos',
            exp='lpSum({}_selfproduction_rate_pos[t] for t in time.I)'
                .format(self.name),
            weight=weight, parent=self)

        self.max_selfproduction_rate_neg = Objective(
            name='max_selfproduction_rate_neg',
            exp='lpSum({}_selfproduction_rate_neg[t] for t in time.I)'
                .format(self.name),
            weight=weight, parent=self)


class Supplier(Consumer, Producer):
    """
    **Description**

        Supplier class inherits from the the class OperatorActor, Consumer
        and Producer. It enables one to model a supplier.
    """

    def __init__(self, name, operated_consumption_unit_list,
                 operated_production_unit_list, verbose=True):
        Consumer.__init__(self, name=name,
                          operated_unit_list=operated_consumption_unit_list,
                          verbose=verbose)
        Producer.__init__(self, name=name,
                          operated_unit_list=operated_production_unit_list,
                          verbose=False)
        self.operated_consumption_unit_list = operated_consumption_unit_list
        self.operated_production_unit_list = operated_production_unit_list
        self.operated_unit_list = operated_consumption_unit_list + \
                                  operated_production_unit_list



#! usr/bin/env python3
#  -*- coding: utf-8 -*-

"""
**This module includes the following display utils:**

    - plot_node_energetic_flows() : enables one to plot the energy flows through
      an EnergyNode
    - plot_energy_mix() : enables one to plot the energy flows connected to a
      node
    - plot_quantity() : enables one to plot easily a Quantitiy
    - plot_quantity_bar() : enables one to plot easily a Quantitiy as a bar
    - sum_quantities_in_quantity() : enables one to to plot several quantities
      in one once the optimisation is done

..
    Copyright 2018 G2Elab / MAGE

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from ...energy.energy_nodes import EnergyNode
from ...energy.units.production_units import ProductionUnit
from ...energy.units.storage_units import StorageUnit
from ...energy.units.consumption_units import ConsumptionUnit
from ..optimisation.elements import Quantity

__docformat__ = "restructuredtext en"


def plot_node_energetic_flows(node):
    """
    **Description**

    This function allows to plot the energy flows through an EnergyNode

    The display is realized :

       - with histograms for production and storage flow
       - with dashed curve for consumption flow

    :param node: EnergyNode
    """
    matplotlib.style.use('seaborn')

    if not isinstance(node, EnergyNode):
        raise TypeError('The element {0} should be an EnergyNode.'.format(node))
    energy_fig = plt.figure()

    print(
        "\nPreparing to plot the energetic flows through the node {0}.".format(
            node.name))

    cons = energy_fig.add_subplot(111)
    previous_energy_production = [0] * node.time.LEN  # Build the base to the
    # bar graph

    # For all production units connected to the node
    flows = node.get_flows
    for flow in flows:
        print(("\tAdd power from {0}.".format(flow.parent.name)))

        label = flow.parent.name
        parent = flow.parent

        # Get the power profiles
        if isinstance(flow.value, list):
            energy_flow = flow.value
        elif isinstance(flow.value, dict):
            energy_flow = list(flow.value.values())

        # If production or import
        if isinstance(parent, ProductionUnit) or node.is_import_flow(flow):
            try:
                plt.bar(node.time.I * node.time.DT, energy_flow,
                        width=0.8 * node.time.DT,
                        bottom=[previous_energy_production[t] for t
                                in range(node.time.LEN)],
                        label=label)

            except IndexError:
                plt.bar(node.time.I * node.time.DT, energy_flow,
                        width=0.8 * node.time.DT,
                        bottom=[previous_energy_production[t] for t
                                in range(node.time.LEN)],
                        label=label)

            for t in range(node.time.LEN):
                previous_energy_production[t] += energy_flow[t]

        # If consumption or export
        elif isinstance(parent, ConsumptionUnit) or node.is_export_flow(flow):
            cons.plot(node.time.I * node.time.DT, energy_flow, marker='.',
                      label=label,
                      linewidth=0.8, linestyle='--')

        # If storage
        elif isinstance(parent, StorageUnit):
            energy_discharge = list(max(-1 * p, 0) for p in energy_flow)
            energy_charge = list(min(-1 * p, 0) for p in energy_flow)

            # Plot the discharge
            try:
                plt.bar(node.time.I * node.time.DT, energy_discharge,
                        width=0.8 * node.time.DT,
                        bottom=[previous_energy_production[t] for t
                                in range(node.time.LEN)],
                        label=label + 'discharge')

            except IndexError:
                plt.bar(node.time.I * node.time.DT, energy_discharge,
                        width=0.8 * node.time.DT,
                        bottom=[previous_energy_production[t] for t
                                in range(node.time.LEN)],
                        label=label)

            for t in range(node.time.LEN):
                previous_energy_production[t] += energy_discharge[t]

            # Plot the charge
            plt.bar(node.time.I * node.time.DT, energy_charge,
                    width=0.8 * node.time.DT,
                    label=label + ' charge')

    # for flow in node._exports:
    #     export = list(flow.value.values())
    #     cons.plot(export, marker='x', label='export')

    plt.title(
        'Power flow for the units connected to the node {0}'.format(node.name))
    if node.time.DT == 1:
        plt.xlabel('Time (hours)')
        plt.ylabel("Hourly mean power (kW)")
    elif node.time.DT == 24:
        plt.xlabel('Time (days)')
        plt.ylabel("Daily total energy (kWh)")

    plt.legend()
    return plt


def plot_energy_mix(node):
    if not isinstance(node, EnergyNode):
        raise TypeError('The element {0} should be an EnergyNode.'.format(node))
    data = []
    legend = []

    _, ax = plt.subplots()
    plt.axis('equal')  # Should be a circle

    flows = node.get_flows
    for flow in flows:
        if isinstance(flow.parent, ProductionUnit):
            energy_prod = sum(flow.value.values())
            data.append(energy_prod)
            legend.append(flow.parent.name)

    ax.pie(data, labels=legend, autopct='%1.1f%%')

    plt.title('Energy mix of the node {0}'.format(node.name))
    plt.legend()


def plot_quantity(time, quantity, fig=None, ax=None, color=None, label=None,
                  title=None):
    """
    **Description**

        Function that plots a OMEGAlpes.general.optimisation.elements.Quantity

    **Parameters**

        - time: TimeUnit for the studied horizon as defined in general.time
        - quantity: OMEGAlpes.general.optimisation.elements.Quantity
        - fig: Figure as defined in matplotlib.pyplot.Figure
        - ax: axes as defined in matplotlib.pyplot.Axes
        - color: color of the plot
        - label: label for the quantity
        - title: title of the plot


    **Returns**

        - arg1 the matplotlib.pyplot.Figure handle object
        - arg2 the matplotlib.pyplot.Axes handle object
        - arg3 the matplotlib.pyplot.Line2D handle object

    """

    v = getattr(quantity, 'value')

    if fig is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    if isinstance(fig, plt.Figure):
        if ax is None:
            ax = fig.add_subplot(111)
        elif not isinstance(ax, plt.Axes):
            raise ValueError(
                'ax should be either NoneType or matplotlib.pyplot.Axes')

    if isinstance(v, (list, np.ndarray)):
        try:
            ld = ax.plot(time.I * time.DT, v, 'x-', color=color,
                         label=label)
            fig.canvas.draw()

        except ValueError:
            ld = ax.plot(np.arange(0, len(v) * time.DT, time.DT), v, 'x-',
                         color=color, label=label)

    elif isinstance(v, dict):
        ind = list(v.keys())
        val = list(v.values())
        lst = np.array([ind, val]).transpose()
        # lst = np.array([ind, val], ndmin=2)
        lst = lst[lst[:, 0].argsort()]
        ind = lst[:, 0]
        val = lst[:, 1]
        ld = ax.plot(ind * time.DT, val, 'x-', color=color, label=label)

    else:
        raise TypeError(
            'Type {0} is not supported for the plot'.format(
                type(v)))

    if title is None:
        plt.title('Evolution of {0} over the studied period'.format(
            quantity.name))
    elif isinstance(title, str):
        plt.title(title)
    else:
        TypeError('title should be either None or a string')

    return ld, ax, fig


def plot_quantity_bar(time, quantity, fig=None, ax=None, color=None,
                      label=None, title=None):
    """
    **Description**

        Function that plots a OMEGALPES.general.optimisation.elements.Quantity
        as a bar

    **Attributes**

        - quantity is the OMEGALPES.general.optimisation.elements.Quantity
        - fig could be None, a matplotlib.pyplot.Figure or Axes for multiple
          plots

    **Returns**

        - arg1 the matplotlib.pyplot.Figure handle object
        - arg2 the matplotlib.pyplot.Axes handle object
        - arg3 the matplotlib.pyplot.Line2D handle object

    """

    v = getattr(quantity, 'value')

    if fig is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    if isinstance(fig, plt.Figure):
        if ax is None:
            ax = fig.add_subplot(111)
        elif not isinstance(ax, plt.Axes):
            raise ValueError(
                'ax should be either NoneType or matplotlib.pyplot.Axes')

    if isinstance(v, (list, np.ndarray)):
        try:
            ld = ax.bar(time.I * time.DT, v, color=color, label=label)
            fig.canvas.draw()

        except ValueError:
            ld = ax.bar(np.arange(0, len(v) * time.DT, time.DT), v,
                        color=color, label=label)

    elif isinstance(v, dict):
        ind = list(v.keys())
        val = list(v.values())
        lst = np.array([ind, val]).transpose()
        # lst = np.array([ind, val], ndmin=2)
        lst = lst[lst[:, 0].argsort()]
        ind = lst[:, 0]
        val = lst[:, 1]
        ld = ax.bar(ind * time.DT, val, color=color, label=label)

    else:
        raise TypeError(
            'Type {0} is not supported for the plot'.format(
                type(v)))

    if title is None:
        plt.title('Evolution of {0} over the studied period'.format(
            quantity.name))
    elif isinstance(title, str):
        plt.title(title)
    else:
        TypeError('title should be either None or a string')

    return ld, ax, fig


def sum_quantities_in_quantity(quantities_list=[],
                               tot_quantity_name='sum_quantity'):
    """
    **Description**

        Function that creates a new quantity gathering several values of
        quantities
        Should be used in order to plot several quantities in one once the
        optimisation is done

    **Attributes**

        * quantities : a list of Quantities
          (OMEGALPES.general.optimisation.elements.Quantity)
        * tot_quantity_name : string : name of the new quantity

    **Returns**

        - tot_quantity : the new quantity created and filled
    """
    tot_q_value = [0] * quantities_list[0].vlen
    qunit = quantities_list[0].unit

    for q in quantities_list:
        if isinstance(q, Quantity):
            if q.unit == qunit:
                for i in range(q.vlen):
                    tot_q_value[i] = tot_q_value[i] + q.value[i]
            else:
                TypeError("The unit {0} of the quantity {1} is not the same "
                          "as the first quantity {2} which unit is {3}"
                          "".format(q.unit, q.name, quantities_list[0].name,
                                    qunit))
        else:
            TypeError("{0} should be a Quantity and is a {1}".format(q.name,
                                                                     q.__class__))

    tot_quantity = Quantity(value=tot_q_value, name=tot_quantity_name,
                            description='create a new quantity gathering '
                                        'various quantities')
    return tot_quantity
